from typing import List, Optional
from datetime import datetime

# for sqlalchemy 2.0 ORM
# see https://docs.sqlalchemy.org/en/20/orm/declarative_config.html
# and https://docs.sqlalchemy.org/en/20/orm/declarative_tables.html#orm-declarative-table

from sqlalchemy import ForeignKey, Index  # pylint: disable=import-error
from sqlalchemy import String  # pylint: disable=import-error
from sqlalchemy import Integer  # pylint: disable=import-error
from sqlalchemy import DateTime  # pylint: disable=import-error
from sqlalchemy import JSON  # pylint: disable=import-error
from sqlalchemy.orm import backref  # pylint: disable=import-error
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error
from sqlalchemy.orm import relationship  # pylint: disable=import-error
from sqlalchemy.orm import object_session  # pylint: disable=import-error
from sqlalchemy import inspect  # pylint: disable=import-error

from timelink.kleio.utilities import (
    kleio_escape,
    get_extra_info_from_obs,
    render_with_extra_info,
    format_timelink_date as ftld,
)
from .base_class import Base


class Entity(Base):
    """ORM Model root of the object hierarchy.

    All entities in a Timelink/MHK database have an entry in this table.
    Each entity is associated with a class that allow access to a
    specialization table with more columns for that class.

    This corresponds to the model described as "Joined Table Inheritance"
    in sqlalchemy (see https://docs.sqlalchemy.org/en/14/orm/inheritance.html)

    TODO: specialize TemporalEntity to implement https://github.com/time-link/timelink-kleio/issues/1
            Acts, Sources, Attributes and Relations are TemporalEntities
    """

    __tablename__ = "entities"
    __allow_unmapped__ = True

    #: str: unique identifier for the entity
    id: Mapped[str] = mapped_column(primary_key=True)
    #: str: name of the class. Links to pom_som_mapper class
    # TODO: https://docs.sqlalchemy.org/en/20/orm/declarative_config.html
    pom_class: Mapped[str] = mapped_column(
        "class",
        String,
        # ForeignKey('classes.id', name='fk_entity_class', use_alter=True),
        index=True,
    )
    #: str: id of the entity inside which this occurred.
    inside: Mapped[Optional[str]] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), index=True
    )
    # id of the source from which this entity was extracted
    the_source: Mapped[Optional[str]] = mapped_column(String, nullable=True)
    #: int: sequential order of this entity in the source
    the_order = mapped_column(Integer, nullable=True)
    #: int: the nesting level of this entity in the source
    the_level = mapped_column(Integer, nullable=True)
    #: int: line in which the entity occurred in the source
    the_line = mapped_column(Integer, nullable=True)
    #: str: name of the kleio group that produced this entity
    groupname = mapped_column(String, index=True, nullable=True)
    # extra_info a JSON field with extra information about the entity
    # For each column of the entity, the value is a dictionary with the following keys:
    # 'class': {'db_column_name': 'class',
    #         'kleio_element_class': 'class',
    #         'kleio_element_name': 'class'},
    # 'date': {'db_column_name': 'the_date',
    #         'kleio_element_class': 'date',
    #         'kleio_element_name': 'date'},
    # 'entity': {'db_column_name': 'entity',
    #             'kleio_element_class': 'entity',
    #             'kleio_element_name': 'entity'},
    # 'id': {'db_column_name': 'id',
    #         'kleio_element_class': 'id',
    #         'kleio_element_name': 'id'},
    # 'type': {'db_column_name': 'the_type',
    #         'kleio_element_class': 'type',
    #         'kleio_element_name': 'tipo'},
    # 'value': {'comment': '@wikidata:Q1171',
    #         'db_column_name': 'the_value',
    #         'kleio_element_class': 'value',
    #         'kleio_element_name': 'valor'}}
    extra_info = mapped_column(JSON, nullable=True)
    #: datetime: when this entity was updated in the database
    updated = mapped_column(DateTime, default=datetime.utcnow, index=True)
    #: datetime: when this entity was added to the full text index
    indexed = mapped_column(DateTime, index=True, nullable=True)

    # This is defined in attribute.py
    # Entity.attributes = relationship(
    #   "Attribute", foreign_keys=[Attribute.entity], back_populates="the_entity"
    #   )
    attributes = None

    # These are defined in relation.py
    # rels_in = relationship("Relation", back_populates="dest")
    # rels_out = relationship("Relation", back_populates="org")

    rels_in = None
    rels_out = None

    # This is defined in REntity.py
    links = relationship(
        "Link", back_populates="entity_rel", cascade="all, delete-orphan"
    )

    # this based on
    # https://stackoverflow.com/questions/28843254
    #: list(Entity): list of Entity objects contained in this entity
    contains: Mapped[List["Entity"]] = relationship(
        "Entity",
        backref=backref("contained_by", remote_side="Entity.id"),
        cascade="all",
    )

    # group_models = contains the correspondence between groupname and ORM class
    # Ensure mappings are up to date
    # see self.ensure_mappings()
    # see PomSomMapper.group_to_entity(self)
    group_models: dict = {}

    # maps group elements to columns. updated by PomSomMapper.kgroup_to_entity
    # this is a dictionary of dictionaries: first key the group name, second key the element name
    group_elements_to_columns: dict[str, dict[str, str]] = {"entity": {"id": "id"}}

    # see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    # To handle non mapped pom_class
    #      see https://github.com/sqlalchemy/sqlalchemy/issues/5445
    #
    #    __mapper_args__ = {
    #       "polymorphic_identity": "entity",
    #    "polymorphic_on": case(
    #        [(type.in_(["parent", "child"]), type)], else_="entity"
    #    ),
    #
    #  This defines what mappings do exist
    # [aclass.__mapper_args__['polymorphic_identity']
    #              for aclass in Entity.__subclasses__()]

    __mapper_args__ = {
        "polymorphic_identity": "entity",
        "polymorphic_on": pom_class,
    }

    @classmethod
    def get_subclasses(cls):
        """Get the subclasses of Entity"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def get_orm_models(cls):
        """Currently defined ORM classes that extend Entity
        (including Entity itself)


        Returns:
            list: List of ORM classes
        """
        sc = list(Entity.get_subclasses())
        return [Entity] + sc

    @classmethod
    def get_som_mapper_ids(cls):
        """Ids of PomSomMapper references by orm classes

        Returns:
            List[str]: List of strings
        """
        return [
            aclass.__mapper_args__["polymorphic_identity"]
            for aclass in Entity.get_orm_models()
        ]

    @classmethod
    def get_tables_to_orm_as_dict(cls):
        """
        Return a dict with table name as key and ORM class as value
        """
        return {
            ormclass.__mapper__.local_table.name: ormclass
            for ormclass in Entity.get_orm_models()
        }

    @classmethod
    def get_orm_table_names(cls):
        """ Return the names of tables associated with ORM models"""
        return [str(table) for table in cls.get_tables_to_orm_as_dict().keys()]

    @classmethod
    def get_group_models(cls) -> dict:
        """ Return a dictionary of group models

        The keys are group names and the values are ORM
        Models.
        """
        return cls.group_models

    @classmethod
    def get_groups_for_orm(cls, orm: str):
        """ get the list of groups that map to this model

        Args:
           orm (str):  id of a orm model
        """
        return [group for (group, orm)
                in cls.get_group_models().items()
                if orm.id == str]

    @classmethod
    def get_orm_for_group(cls, groupname: str):
        """
        Entity.get_orm_for_group("acts")

        will return the ORM class handling a given group
        """
        meta_group_models = cls.get_group_models()
        if meta_group_models is not None:
            return meta_group_models.get(groupname, None)
        else:
            raise ValueError(f"{groupname} is not associated with any ORM Model class")

    @classmethod
    def set_orm_for_group(cls, groupname: str, ormclass):
        cls.group_models[groupname] = ormclass

    @classmethod
    def clear_group_models_cache(cls):
        """Clear cached association between groups and ormclass"""
        cls.group_models = dict()

    @classmethod
    def get_som_mapper_to_orm_as_dict(cls):
        """
        Return a dict with pom_class id as key and ORM class as value
        """
        sc = Entity.get_orm_models()
        return {ormclass.__mapper__.polymorphic_identity: ormclass for ormclass in sc}

    @classmethod
    def get_orm_for_table(cls, table: String):
        """
        Entity.get_orm_for_table("acts")

        will return the ORM class handling the "acts" table
        """
        return cls.get_tables_to_orm_as_dict().get(table, None)

    @classmethod
    def get_orm_for_pom_class(cls, pom_class: str):
        """
        Entity.get_orm_for_pom_class("act")

        will return the ORM class corresponding to the pom_class "act"
        """
        return cls.get_som_mapper_to_orm_as_dict().get(pom_class, None)

    @classmethod
    def get_entity(cls, eid: str, session=None):
        """
        Get an Entity from the database. The object returned
        will be of the ORM class defined by mappings.
        :param id: id of the entity
        :param session: current session
        :return: an Entity object of the proper class for the mapping
        """
        entity = session.get(Entity, eid)
        if entity is not None:
            if entity.pom_class != "entity":
                orm_class = Entity.get_orm_for_pom_class(entity.pom_class)
                object_for_id = session.get(orm_class, eid)
                return object_for_id
            else:
                return entity
        else:
            return None

    def get_column_for_element(self, element: str):
        """Get the column name for a group element"""
        self.update_group_elements_to_columns
        return Entity.group_elements_to_columns.get(self.groupname, {}).get(
            element, None
        )

    def get_element_for_column(self, column: str):
        """Get the kleio element name for a column"""

        self.update_group_elements_to_columns()
        extra_info = getattr(self, "extra_info", None)
        element_name = None
        if extra_info is not None:
            col_extra_info = extra_info.get(column, None)
            if col_extra_info is not None:
                element_name = col_extra_info.get("kleio_element_name", None)
        if element_name is not None:
            return element_name
        # if we didn't find it in extra_info, look in the group_elements_to_columns
        for element, column_name in Entity.group_elements_to_columns.get(
            self.groupname, {}
        ).items():
            if column_name == column:
                return element
        return None

    def update_group_elements_to_columns(self):
        """Update the element to column mappings for the Group of this Entity"""
        # Only do this if we don't have any mappings for this Entity
        if (
            Entity.group_elements_to_columns.get(self.groupname, None) is None
            or len(  # noqa: W503
                Entity.group_elements_to_columns.get(self.groupname, {})
            )
            == 0  # noqa: W503
        ):
            session = object_session(self)
            if session is None:
                raise ValueError(
                    "Entity must be in a session to update element to columns" "mappings"
                )
                # when a group is added programatically, the element to column mappings
                # the pom_class can be None. In this case, we use the group name
                # as the pom_class
            if self.pom_class is None:
                self.pom_class = self.groupname
                # get the PomSomMapper for this group
            psm = session.get(Entity, self.pom_class)  # <- PomSomMapper
            if psm is None:
                raise ValueError(f"No PomSomMapper found for this group {self.groupname}")
            # get the element to column mappings
            elcol_map: dict = Entity.group_elements_to_columns.get(self.groupname, {})
            # update the element to column mappings
            for cattr in psm.class_attributes:  # noqa
                elcol_map[cattr.name] = cattr.colname
            # put it back in the dictionary
            Entity.group_elements_to_columns[self.groupname] = elcol_map

    def get_element_names(self):
        """Get the names of the elements for this entity"""
        self.update_group_elements_to_columns()
        # it is better to try to get from the extra_info
        if getattr(self, "extra_info", None) is not None:
            columns = self.extra_info.keys()
            element_names = [self.get_element_for_column(col) for col in columns]
            return element_names
        # if not, get from the group_elements_to_columns
        return Entity.group_elements_to_columns.get(self.groupname, {}).keys()

    def add_attribute(self, attribute):
        """Add an attribute to the entity"""
        if attribute.inside is None:
            attribute.inside = self.id
        self.attributes.append(attribute)

    def add_relation(self, relation):
        """Add a relation to the entity"""
        if relation.inside is None:
            relation.inside = self.id
        if relation.origin is None or self.id == relation.origin:
            # no origin, this is the origin
            relation.origin = self.id
            self.rels_out.append(relation)
        elif relation.destination is None or self.id == relation.destination:
            # no destination, this is the destination
            relation.destination = self.id
            self.rels_in.append(relation)
        else:
            raise ValueError("Relation does not belong to this entity")

    def with_extra_info(self):
        """Returns a copy of the entity with field values rendered with extra information

        Extra_info is current stored as json in the obs field of the entity, or
        in the extra_info field of the entity. The extra_info field is a json field.
        """
        # the the current session
        session = object_session(self)
        if session is None:
            raise ValueError("Entity must be in a session to get extra info")
        # create a new object of the same class
        new_entity = self.__class__()
        mapper = inspect(type(self))
        field_to_column = {
            col.key: col.columns[0].name for col in list(mapper.column_attrs)
        }
        obs, extra_info = self.get_extra_info()
        for name, __column in field_to_column.items():
            nvalue = render_with_extra_info(name, getattr(self, name), extra_info)
            setattr(new_entity, name, nvalue)

        setattr(new_entity, "obs", obs)  # noqa

        return new_entity

    def get_extra_info(self) -> tuple[str, dict]:
        """Return a dictionatry with extra information about this entity or None

        This method checks if there is extra information about the entity.
        This is needed because in the source oriented notation each
        value can have a comment and the original wording of the value,
        and also multiple values for a single element/field.

        The way this is managed in the relational database (person oriented model)
        is to store the information as a Json object in the 'extra_info' field.
        A legacy pattern was to store the information in the 'obs' field.

        if an entity has an 'obs' field and that field
        contains the string 'extra_info:' then the rest
        of the field is considered a json representation of
        extra information about the entity. Extra information can
        be comments and original wording of field values.

        This method returns a tuple with the extra information
        as dict and, in the case that the obs field
        contains extra information, a new obs string
        (withou the extra_info part).


        Currently the dictionary has the following structure:

            {'field_name':
                {'comment': text_of_comment,
                 'original': original_wording}}


        """
        extra_info = {}
        if hasattr(self, "obs"):
            obs = getattr(self, "obs", None)
        else:
            obs = None
        if obs is not None and "extra_info:" in obs:
            obs, extra_info = get_extra_info_from_obs(obs)

        if hasattr(self, "extra_info") and self.extra_info is not None:
            extra_info = getattr(self, "extra_info", None)
        return obs, extra_info

    # render related methods
    def description_elements(self):
        """Return a list of elements to be used in the description of the entity"""
        return ["name", "description", "desc", "title", "id"]

    def get_description(self, default=None):
        """Return a descritive name for the entity"""
        desc = None
        for element in self.description_elements():
            if hasattr(self, element):
                desc = getattr(self, element)
                break
        if desc is None:
            desc = default
        return desc

    def __repr__(self):
        return (
            f'Entity(id="{self.id}", '
            f'pom_class="{self.pom_class}", '
            f'inside="{self.inside}", '
            f"the_source={self.the_source}, "
            f"the_order={self.the_order}, "
            f"the_level={self.the_level}, "
            f"the_line={self.the_line}, "
            f'groupname="{self.groupname}", '
            f"updated={self.updated}, "
            f"indexed={self.indexed})"
        )

    def __str__(self):
        groupname = self.groupname
        if groupname is None:
            groupname = self.pom_class
        if groupname is None:
            groupname = "entity"
        desc = self.get_description()
        if desc is not None:
            r = f"{groupname}${desc}"
        else:
            r = f"{groupname}${self.id}"
        els = self.get_element_names()
        dels = self.description_elements()
        show = sorted(list(set(els) - set(dels)))
        if "obs" in show:
            show.remove("obs")
            show.append("obs")
        for element in show:
            r = f"{r}{self.for_kleio(element, prefix='/', named=True, skip_if_empty=True)}"
        r = r + self.render_id()
        return r

    def render_id(self):
        """Return the id of the entity in a kleio format

        Does not return the id if it starts with an underscore"""
        if self.id[:1] == "_":
            return ""
        else:
            return f"/id={self.id}"

    def dated_bio(self) -> dict:
        """Return the atributes and relations of the entity grouped by date

        Returns a dictionary with the date as key and a list of attributes and relations as value

        """
        bio = dict()

        if self.rels_in is not None:
            for rel in self.rels_in:
                date = rel.the_date
                this_date_list = bio.get(date, [])
                this_date_list.append(rel)
                bio[date] = this_date_list
        if self.rels_out is not None:
            for rel in self.rels_out:
                date = rel.the_date
                this_date_list = bio.get(date, [])
                this_date_list.append(rel)
                bio[date] = this_date_list
        if self.attributes is not None:
            for attr in self.attributes:
                date = attr.the_date
                this_date_list = bio.get(date, [])
                this_date_list.append(attr)
                bio[date] = this_date_list
        return bio

    def is_inbound_relation(self, relation):
        """Check if the relation is inbound to this entity.

        Override in real entities or other special cases
        """
        return relation.destination == self.id

    def for_kleio(self, element: str, named=False, prefix="", skip_if_empty=False):
        """Render a specific attribute/column/element of the entity for kleio.

        Check which attribute is requested and return the value
        Checks for extra information for this attribute.

        Extra information can be "comment" or "original"

        if the attribute has a comment, #comment is appended to the value
        if the attribute has an original wording, %original is appended to the value

        """
        # check if the element is a mapped kleio group element
        attr_name = self.get_column_for_element(element)
        if attr_name is None and hasattr(self, element):
            # if not take it as a direct attribute
            # this means the element is a column name like "the_date"
            # and not a group element like "date"
            # but we still the element name to check
            # for extra information
            attr_name = element
            element = self.get_element_for_column(attr_name)

        if hasattr(self, attr_name):
            attr = getattr(self, attr_name)
            if attr is None:
                attr = ""
            else:
                attr = str(attr)  # we need this for enumerations and numbers
            # if the lement is obs that is can cointain extra information
            # that we need to extract and remove from the fied
            obs, extra_info = self.get_extra_info()
            if element == "obs":  # if the element is obs, replace without extra_info
                if obs is not None:
                    attr = obs.strip()
                else:
                    attr = ""
            attr = kleio_escape(attr)
            if element == "date":  # TODO this should test element class not name
                # if the date a just composed of digits try to format it
                if attr.isdigit():
                    attr = ftld(attr)
            if extra_info is not None and element in extra_info:
                extras = extra_info.get(element, {})
                element_comment = extras.get("comment", None)
                element_original = extras.get("original", None)
                if element_comment is not None:
                    attr = f"{attr}#{kleio_escape(element_comment)}"
                if element_original is not None:
                    attr = f"{attr}%{kleio_escape(element_original)}"
            if skip_if_empty and attr == "":
                return ""
            else:
                return (f"{prefix}{element}=" if named else "") + attr
        else:
            return ""

    def to_kleio(
        self, self_string=None, show_contained=True, ident="", ident_inc="  ", **kwargs
    ):
        """conver the entity to a kleio string

        Args:
            self_string: the string to be used to represent the entity
            show_contained: if True, contained entities are also converted to kleio
            ident: initial identation
            ident_inc: identation increment
            kwargs: additional arguments to be passed to contained entities"""
        if self_string is None:
            s = f"{ident}{str(self)}"
        else:
            s = f"{ident}{self_string}"

        contained_entities = list(
            set(self.contains)
            - set(self.rels_in)  # noqa: W503
            - set(self.rels_out)  # noqa: W503
            - set(self.attributes)  # noqa: W503
        )
        bio = self.dated_bio()
        sorted_keys = sorted(bio.keys())
        for date in sorted_keys:
            date_list = bio[date]
            for bio_item in date_list:
                bio_item_xi = bio_item
                if bio_item.pom_class == "relation":

                    kwargs["outgoing"] = not self.is_inbound_relation(bio_item)
                    if self.is_inbound_relation(bio_item):
                        if bio_item.the_type == "function-in-act":
                            # we don't render inbound function-in-act relations
                            # because they are redundant with contained entities
                            continue

                bio_itemk = bio_item_xi.to_kleio(
                    ident=ident + ident_inc, ident_inc=ident_inc, **kwargs
                )

                if bio_itemk != "":
                    s = f"{s}\n{bio_itemk}"

        # sort by the_order
        if show_contained and contained_entities is not None:
            contained_entities.sort(
                key=lambda x: x.the_order if x.the_order is not None else 999999
            )
            for inner in contained_entities:
                innerk = inner.to_kleio(
                    ident=ident + ident_inc, ident_inc=ident_inc, **kwargs
                )
                if innerk != "":
                    s = f"{s}\n{innerk}"
        return s


#  Add composite index to speed up
Index("entities_class_group", Entity.pom_class, Entity.groupname)



from typing import Optional, List
from xml.sax import handler, SAXParseException

from timelink.api.models import PomSomMapper as PomSomMapperTL, PomClassAttributes as PomClassAttributesTL
from timelink.kleio.groups import KGroup, KElement
from timelink.kleio.kleio_handler import KleioContext
from timelink.mhk.models import PomSomMapper as PomSomMapperMHK, PomClassAttributes as PomClassAttributesMHK


class SaxHandler(handler.ContentHandler):
    """SAX Handler for Kleio XML files"""

    _context: KleioContext
    _current_class: Optional[PomSomMapperTL | PomSomMapperMHK]
    _current_class_attrs: List[PomClassAttributesTL | PomClassAttributesMHK]
    _current_group: Optional[KGroup]
    _current_element: Optional[KElement]
    _current_entry: str

    def __init__(self, kleio_handler):
        handler.ContentHandler.__init__(self)
        self._kleio_handler = kleio_handler

    def startDocument(self):
        self._context = KleioContext.START
        self._current_class = None
        self._current_class_attrs = []
        self._current_group = None
        self._current_element = None
        self._current_entry = ""

    def endDocument(self):
        self._kleio_handler.endKleioFile()

    def startElement(self, name: str, attrs):
        #  https://stackoverflow.com/questions/15477363/xml-sax-parser-and-line-numbers-etc
        loc = self._locator
        ename = name.upper()
        if ename == "KLEIO":
            self._kleio_handler.newKleioFile(attrs)
            self._context = KleioContext.KLEIO
        elif ename == "CLASS":
            # <CLASS NAME="source" SUPER="entity"
            #       TABLE="sources" GROUP="fonte">
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    exception=None,
                    locator=loc,
                    msg="CLASS out of context, should be inside KLEIO element",
                )
            self._current_class = self._kleio_handler.pom_som_mapper(
                id=attrs["NAME"],
                super_class=attrs["SUPER"],
                table_name=attrs["TABLE"],
                group_name=attrs["GROUP"],
            )
            self._current_class_attrs = []
            self._context = KleioContext.CLASS

        elif ename == "RELATION":
            # <RELATION ID="sa-deh-joao-barradas-rela6"
            #   ORG="sa-deh-joao-barradas-per1-11"
            #   DEST="sa-deh-matteo-ricci"
            #   TYPE="META" VALUE="same_as"/>
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    "RELATION out of context, should be inside KLEIO element",
                    None,
                    loc,
                )
            self._kleio_handler.newRelation(attrs)

        elif ename == "ATTRIBUTE":
            #  <ATTRIBUTE NAME="id"
            #       COLUMN="id"
            #       CLASS="id"
            #       TYPE="varchar"
            #       SIZE="64"
            #       PRECISION="0"
            #       PKEY="1" >
            #   </ATTRIBUTE>
            if self._context != KleioContext.CLASS:
                raise SAXParseException(
                    "ATTRIBUTE out of context, should be inside CLASS element",
                    None,
                    loc,
                )
            new_attribute = self._kleio_handler.pom_class_attributes()
            new_attribute.the_class = self._current_class.id
            new_attribute.name = attrs["NAME"]
            new_attribute.colname = attrs["COLUMN"]
            new_attribute.colclass = attrs["CLASS"]
            new_attribute.coltype = attrs["TYPE"]
            new_attribute.colsize = int(attrs["SIZE"])
            new_attribute.colprecision = int(attrs["PRECISION"])
            new_attribute.pkey = int(attrs["PKEY"])

            self._current_class_attrs.append(new_attribute)

        elif ename == "GROUP":
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    "GROUP out of context, should be inside KLEIO element", None, loc
                )
            # <GROUP ID="coja-rol-1841" NAME="fonte"
            #        CLASS="source" ORDER="1" LEVEL="1" LINE="4">
            gid = attrs["ID"]
            name = attrs["NAME"]
            the_class = attrs["CLASS"]
            order = attrs["ORDER"]
            level = attrs["LEVEL"]
            line = attrs["LINE"]

            # Mal [? porquê?]
            self._current_group = KGroup()
            self._current_group.id = gid
            self._current_group._name = name
            self._current_group._pom_class_id = the_class
            self._current_group._order = int(order)
            self._current_group.level = int(level)
            self._current_group.line = int(line)
            self._context = KleioContext.GROUP

        elif ename == "ELEMENT":
            if self._context != KleioContext.GROUP:
                raise SAXParseException(
                    "ELEMENT out of context, should be inside GROUP element",
                    exception=None,
                    locator=loc,
                )
            # <ELEMENT NAME="id" CLASS="id">
            #       <core>r1775-f1-per1</core></ELEMENT>

            elname = attrs["NAME"]
            self._current_group.allow_as_element(elname)
            self._current_element = KElement(elname, None)
            self._current_element.element_class = attrs["CLASS"]
            self._context = KleioContext.ELEMENT

        elif ename == "CORE":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "CORE out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.CORE
            self._current_entry = ""

        elif ename == "COMMENT":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "COMMENT out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.COMMENT
            self._current_entry = ""

        elif ename == "ORIGINAL":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "ORIGINAL out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.ORIGINAL
            self._current_entry = ""

        elif ename == "RELATION":
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    "GROUP out of context, should be inside KLEIO element",
                    exception=None,
                    locator=loc,
                )
            self._kleio_handler.newRelation(attrs)
        else:
            raise SAXParseException(
                f"Unexpected element in Kleio file: {ename}",
                exception=None,
                locator=loc,
            )

    def endElement(self, name):
        ename = name.upper()

        if ename == "GROUP":
            self._kleio_handler.newGroup(self._current_group)
            self._context = KleioContext.KLEIO
            self._current_group = None

        elif ename == "CLASS":
            self._kleio_handler.newClass(self._current_class, self._current_class_attrs)
            self._context = KleioContext.KLEIO

        elif ename == "ATTRIBUTE":
            self._context = KleioContext.CLASS

        elif ename == "ELEMENT":
            self._current_group[self._current_element.name] = self._current_element

            self._context = KleioContext.GROUP

        elif ename in ["CORE", "COMMENT", "ORIGINAL"]:
            if ename == "CORE":
                if self._current_element.core is None:
                    self._current_element.core = self._current_entry
            if ename == "COMMENT":
                if self._current_element.comment is None:
                    self._current_element.comment = self._current_entry
            if ename == "ORIGINAL":
                if self._current_element.original is None:
                    self._current_element.original = self._current_entry
            self._context = KleioContext.ELEMENT

    def characters(self, content):
        self._current_entry = self._current_entry + content

    def ignorableWhitespace(self, whitespace):
        pass

    def processingInstruction(self, target, data):
        pass



# This is a file for the "migrations" folder in timelink.
# It is used to hold the alembic related files and configurations.
# These allow for the update of databases when the models from within
# the timelink package without user intervention.
#
# this file and the techniques used are based on the following SO answer:
# https://stackoverflow.com/questions/73633063/distribute-alembic-migration-scripts-in-application-package
# This might be generating and error related to "formaters" in some cases.
# Check the following link for more information:
# https://stackoverflow.com/questions/47636747/python3-6-keyerror-formatters
# it might be related to the way we configure the path to the alembic.ini file.
from pathlib import Path

from alembic import command
from alembic.runtime import migration
from alembic.config import Config
from alembic.script import ScriptDirectory
from sqlalchemy import engine
from sqlalchemy import inspect


ROOT_PATH = Path(__file__).parent.parent
ALEMBIC_CFG = Config(ROOT_PATH / "alembic.ini")
ALEMBIC_CFG.set_main_option("script_location", str(ROOT_PATH / "migrations"))


# Auxiliary function for migrations
def column_exists(table_name, column_name, alembic_op):
    """ Check if a column exists

    Args:
        table_name: name of table
        column_name: name of column
        alembic_op:  alembic operations object in use

    Example:


    """

    bind = alembic_op.get_context().bind
    insp = inspect(bind)
    columns = insp.get_columns(table_name)
    return any(c["name"] == column_name for c in columns)


# Naming conventions for foreign keys
# https://alembic.sqlalchemy.org/en/latest/naming.html#naming-conventions
naming_convention = {
    "fk":
    "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
}


def set_db_url(db_url):
    """Set the database URL in the alembic configuration.

    This should be done before running any alembic commands in a specific database
    """
    ALEMBIC_CFG.set_section_option("alembic", "sqlalchemy.url", db_url)


# Unused
def get_versions(base='base', head='heads'):
    """Get the list of versions of the database.

    Args:
        base (str): The base revision to start from.
        head (str): The head revision to stop at.
    """
    script = ScriptDirectory.from_config(ALEMBIC_CFG)
    revisions = list(script.walk_revisions(base, head))
    return revisions


def current(db_url, verbose=False):
    """Get the current revision of the database."""
    set_db_url(db_url)
    command.current(ALEMBIC_CFG, verbose=verbose)


def upgrade(db_url, revision="heads"):
    """Upgrade the database to a given revision (default most recent)."""
    set_db_url(db_url)
    command.upgrade(ALEMBIC_CFG, revision)


def downgrade(db_url, revision):
    """Downgrade the database to the given revision."""
    set_db_url(db_url)
    command.downgrade(ALEMBIC_CFG, revision)


def heads(db_url):
    """ Show heads of database"""
    set_db_url(db_url)
    engine_ = engine.create_engine(db_url)
    with engine_.begin() as connection:
        migration_context = migration.MigrationContext.configure(connection)
        heads = migration_context.get_current_heads()
    return heads


def stamp(db_url, revision: str):
    """ Stamp current revision to a given revision
    Use "heads" to mark the database as up-to-date
    """
    set_db_url(db_url)
    command.stamp(ALEMBIC_CFG, revision)


def history(verbose=False):
    """Show the migration history."""
    command.history(ALEMBIC_CFG, verbose=verbose)


def autogenerate(db_url, message):
    """Autogenerate migration script."""
    set_db_url(db_url)
    command.revision(ALEMBIC_CFG, message=message, autogenerate=True)


def revision(db_url, message, autogenerate=False):
    """Create a new revision."""
    set_db_url(db_url)
    command.revision(ALEMBIC_CFG, message=message, autogenerate=autogenerate)



"""This module provides utility functions for interacting with the Timelink database,
including generating random passwords, retrieving database passwords from the environment,
and determining the import status of Kleio files.

Functions:
    random_password() -> str:
        Generate a random password consisting of 10 ASCII letters.

    get_db_password() -> str:
        Retrieve the database password from the environment. If not set, generate a new one,
        set it in the environment, and return it.

    get_import_status(db: TimelinkDatabase, kleio_files: List[KleioFile], match_path=False) -> List[KleioFile]:
        Determine the import status of the provided Kleio files by comparing them with previously
        imported files in the database. Optionally match files by their path instead of their name.

"""
from __future__ import annotations
from datetime import timezone
import os
import random
import string
from typing import List, TYPE_CHECKING

from timelink.api.models.system import KleioImportedFileSchema
from timelink.kleio import KleioFile, import_status_enum

if TYPE_CHECKING:
    from timelink.api.database import TimelinkDatabase


def random_password():
    """Generate a random password"""

    letters = string.ascii_letters
    result_str = "".join(random.choice(letters) for i in range(10))
    return result_str


def get_db_password():
    """Get the database password from the environment.
    If none generated one and set it in the environment."""
    # get password from environment
    db_password = os.environ.get("TIMELINK_DB_PASSWORD")
    if db_password is None:
        db_password = random_password()
        os.environ["TIMELINK_DB_PASSWORD"] = db_password
    return db_password


def get_import_status(
    db: TimelinkDatabase, kleio_files: List[KleioFile], match_path=False
) -> List[KleioFile]:
    """Get the import status of the kleio files.

        The status in returned
        in :attr:`timelink.api.models.system.KleioImportedFileSchema.import_status`

    Args:
        db (TimelinkDatabase): timelink database
        kleio_files (List[KleioFile]): list of kleio files with extra field import_status
        match_path (bool, optional): if True, match the path of the kleio file
                                    with the path of the imported file;
                                    defaults to False.

    Returns:
        List[KleioFile]: list of kleio files with field import_status
    """
    # if kleio_files is a single object wrap in a list
    if isinstance(kleio_files, KleioFile):
        kleio_files = [kleio_files]

    previous_imports: List[KleioImportedFileSchema] = db.get_imported_files()
    if match_path:
        imported_files_dict = {imported.path: imported for imported in previous_imports}
        valid_files_dict = {file.path: file for file in kleio_files}
    else:
        imported_files_dict = {imported.name: imported for imported in previous_imports}
        valid_files_dict = {file.name: file for file in kleio_files}
        if len(valid_files_dict) != len(kleio_files):
            raise ValueError(
                "Some kleio files have the same name. "
                "Use match_path=True to match the full path of the kleio file "
                "with the path of the imported file."
            )

    for path, file in valid_files_dict.items():
        if (
            path not in imported_files_dict
            or file.translated is None  # noqa: W503
        ):
            file.import_status = import_status_enum.N
        elif imported_files_dict[path].imported is None:
            # if a reimport of a previous imported file fails its
            # imported date will be None and it needs reimport
            file.import_status = import_status_enum.N
        elif file.translated > imported_files_dict[path].imported.replace(
            tzinfo=timezone.utc
        ):
            file.import_status = import_status_enum.U
        else:
            file.import_errors = imported_files_dict[path].nerrors
            file.import_warnings = imported_files_dict[path].nwarnings
            file.import_error_rpt = imported_files_dict[path].error_rpt
            file.import_warning_rpt = imported_files_dict[path].warning_rpt
            file.imported = imported_files_dict[path].imported
            file.imported_string = imported_files_dict[path].imported_string
            if imported_files_dict[path].nerrors > 0:
                file.import_status = import_status_enum.E
            elif imported_files_dict[path].nwarnings > 0:
                file.import_status = import_status_enum.W
            else:
                file.import_status = import_status_enum.I
    return kleio_files



import logging
import warnings
from typing import Optional, Type, List
import json

# pylint: disable=import-error

from sqlalchemy import Column, String, Integer, ForeignKey, Table, Float, select
from sqlalchemy import inspect
from sqlalchemy.engine import Engine
from sqlalchemy.orm import relationship
from sqlalchemy import exc as sa_exc

from timelink.kleio.groups import KGroup, KElement
from timelink.mhk.models.base_class import Base
from timelink.mhk.models.entity import Entity

logger = logging.getLogger(__name__)


class PomSomMapper(Entity):
    """

    Represents a mapping between a Kleio Group in the
    Source Oriented Model (Som) and a relational database entity
    in the Person Oriented Model (Pom). This class corresponds
    to the table "classes" in a Timelink-MHK database,
    and the associated "class_attributes" table. Together the
    two tables describe a Som-Pom mapping, that define how
    Kleio groups are stored in the relational database.

    This class can generate tables and ORM objects that can
    store a Kleio Group in the database.

    Fields

    :id: name of this PomSomMapper, singular form
    :table_name: name of the table in Pom, plural form
    :group_name: name of Som group that maps to this table
    :super_class: name of PomSom class extended by this one

    For the core kleio groups (source,act,person,object, relation,attribute)
    the mapping information is predefined at database creation time and
    the 'classes' and 'class_attributes' populated accordingly.

    The Kleio translator can provide new mappings for new groups that are
    created for specific sources. The mapping information between new groups
    and the database is embedded in the translator output of new sources in the
    form of data for the tables mapped to the PomSomMapper.

    For a mapping between a Som Group and a Pom table
    to be fully operational it is necessary that:


    1. The tables "classes" and "class_attributes" contain the mapping
       information, normally populated during the initialization process of
       the database (DBSystem.db_init) and updated during import,
       as new sources define new mappings dynamically.

    2. A table for storing the elements of the Som Group.
       This is either a basic core table
       (persons,objects,acts,sources,...) or a table that extends
       a basic core table with extra columns.

    3. An ORM sqlalchemy model that represents the Pom model and joins
       the information of the inheritance hierarchy, by mapping to the necessary
       tables.


    The name of the table for a given group and the correspondence between
    group elements and table columns is handled by the PomSomMapper. If the
    group adds extra information then a new table
    is created  that "extends" an existing core table,
    by what is called a "joined inheritance hierarchy"
    (see https://docs.sqlalchemy.org/en/14/orm/inheritance.html)

    To ensure that all the three dimensions exist in a given context it
    is necessary that:

    1. When creating a new database, the core Pom tables should be created
    using sqlalchemy `metadata.create_all(bind=engine)`. Then the
    tables "classes" and "class_attributes" must be populated with the mapping
    of the core Som and Pom groups and entities.

    See timelink.models.base_mappings.py for the core data for
    bootstrapping of the mapping information.

    Both steps are ensured by DBSystem.init_db

    2. When importing new sources, if a new mapping was generated at the
       translation step, it is necessary first to populate "classes" and
       "class_attributes" with the mapping information, included in the translation
       source, then generate the table for the group information (if it
       does not exist) and the ORM mapping. This step is done by the method
       PomSomMapper.ensure_mapping(session) which must be called during the
       import process.

    3. When connecting to a database created by a legacy version of MHK, it is
       necessary to ensure that all the ORM mappings are created, by examining the
       information in the "classes" table and checking if the ORM mapping and
       table representations exist. This is done by the PomClass.ensure_all_mappings()
       method.

    """

    __tablename__ = "classes"
    __allow_unmapped__ = True

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    table_name = Column(String)
    group_name = Column("group_name", String(32))
    super_class = Column("super", String)
    __mapper_args__ = {
        "polymorphic_identity": "class",
        "inherit_condition": id == Entity.id,
    }
    # stores the ORM mapper for this mapping (source, act,person...)
    orm_class: Entity
    # Class attribute stores current PomSomMappings keyed by id
    pom_classes: dict = {}
    # Stores the association between group names and ORM Models.
    #  See get_orm_for_groupname and issue #53
    group_orm_models: dict = {}

    def ensure_mapping(self, session=None):
        """
        Ensure that a table exists to support
        this SOM Mapping and an ORM class is created
        to represent data of objects from this mapping.

        Checks if there is a table definition
        to support this mapping. If not a new table
        definition is created with information from the
        class attributes.

        A new ORM class is also created for mapping the new
        table. The new ORM class will extend the superclass
        ORM mapping

        """

        if not hasattr(self, "orm_class"):
            self.orm_class = None

        # if we have ensured before return what we found then
        if self.orm_class is not None:
            return self.orm_class

        # if not check if Entity knows about an orm class
        # if so return it and save for next time
        my_orm = Entity.get_orm_for_pom_class(self.id)
        if my_orm is not None:
            self.orm_class = my_orm
            if self.group_name is not None:
                # Store as "static"
                PomSomMapper.group_orm_models[self.group_name] = my_orm
            return self.orm_class

        # we have no ORM mapping for the SomPomMapper
        # First check if we have this table already mapped to some Entity.
        # This might happen if we have different kleio groups mapped to the
        # same table, in order to make the kleio transcripts more readable
        # (it happens frequently with the table 'acts').
        # If so we will reuse the existing Table class
        # It can also happen that while not having a ORM mapping the table
        # can already exist in the database as result of previous imports.
        #
        # Note that non core PomSomMappings and corresponding ORM classes,
        # which are dynamically defined during import, have to be recreated
        # from the database information each time an application runs.
        metadata_obj = type(self).metadata  # pylint: disable=no-member
        pytables = metadata_obj.tables  # these are the tables known to ORM

        dbengine: Engine = session.get_bind()
        insp = inspect(dbengine)
        dbtables = insp.get_table_names()  # these are the ones in the database

        my_table: Table
        if self.table_name in pytables.keys():
            # table is known to ORM we used the Table class there
            my_table = pytables[self.table_name]
        elif self.table_name in dbtables:
            # the table exists in the database, we introspect
            my_table = Table(self.table_name, metadata_obj, autoload_with=dbengine)
            # we need to ensure that the foreign key relation exists with super talbe
            # otherwise the ORM mapping further down will fail.
            if self.super_class not in ["root", "base"]:
                # print("Getting super class " + self.super_class)
                pom_super_class: PomSomMapper = PomSomMapper.get_pom_class(
                    self.super_class, session
                )
                if pom_super_class is not None:
                    super_class_table_id = pom_super_class.table_name + ".id"
                else:
                    message = (
                        "Creating mapping for %s super class %s not found"
                        " Default to entities as super class"
                    )
                    logger.warning(message, self.id, self.super_class)
                    super_class_table_id = "entities.id"
                pytype = my_table.c.id.type
                my_table.append_column(
                    Column(
                        "id",
                        pytype,
                        ForeignKey(super_class_table_id),
                        primary_key=True,
                    ),
                    replace_existing=True,
                )
        else:
            # Table is unknown to ORM mapper and does not exist in the database
            # This is the dynamic part, we create a table with
            # the definition the "class" and "class_attributes" tables
            # fetched by this class
            # NOTE this does not take into account that the columns
            #       may already exist in the super class or up in the
            #       class hierarchy. Not sure if this the way it is supposed
            #       to be or if it is a problem to be dealt with here.
            #       SQLAchemy will rename the columns to avoid conflict
            #       automatically
            my_table = Table(self.table_name, metadata_obj, extend_existing=True)
            cattr: Type["PomClassAttributes"]
            for cattr in self.class_attributes:  # pylint: disable=no-member
                PyType: str
                pom_type = cattr.coltype.lower()
                if pom_type == "varchar":
                    PyType = String(cattr.colsize)
                elif pom_type == "numeric" and cattr.colprecision == 0:
                    PyType = Integer
                elif pom_type == "numeric" and cattr.colprecision > 0:
                    PyType = Float
                else:
                    PyType = String
                # print(f"Inferred python type for {cattr.colname}: ", PyType)

                if cattr.pkey != 0:
                    if self.super_class not in ["root", "base"]:
                        # print("Getting super class " + self.super_class)
                        pom_super_class: PomSomMapper = PomSomMapper.get_pom_class(
                            self.super_class, session
                        )
                        if pom_super_class is not None:
                            super_class_table_id = pom_super_class.table_name + ".id"
                        else:
                            message = (
                                "Creating mapping for %s super class %s not found"
                                " Default to entities as super class"
                            )
                            logger.warning(message, self.id, self.super_class)
                            super_class_table_id = "entities.id"
                        my_table.append_column(
                            Column(
                                cattr.colname,
                                PyType,
                                ForeignKey(super_class_table_id),
                                primary_key=True,
                            ),
                            replace_existing=True,
                        )
                    else:  # we are at the root of the entity hierarchy
                        my_table.append_column(
                            Column(cattr.colname, PyType, primary_key=True),
                            replace_existing=True,
                        )
                else:
                    my_table.append_column(
                        Column(cattr.colname, PyType, primary_key=False),
                        replace_existing=True,
                    )
            my_table.create(session.get_bind())
            session.commit()
            # End of creation of a dynamic mapping and table

        # we know create a new ORM mapping for this PomSomMapper
        super_orm = Entity.get_orm_for_pom_class(self.super_class)
        if super_orm is None:
            super_orm = pom_super_class.ensure_mapping(session)
        props = {
            "__table__": my_table,
            "__mapper_args__": {"polymorphic_identity": self.id},
        }
        try:
            with warnings.catch_warnings():
                # We ignore warning related to duplicate fields in
                # specialized classes (obs normally, but also the_type...)
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)
                my_orm = type(self.id.capitalize(), (super_orm,), props)
        except Exception as exc:   # pylint: disable=broad-except
            logger.ERROR(Exception(f"Could not create ORM mapping for {self.id}"), exc)

        self.orm_class = my_orm
        PomSomMapper.group_orm_models[self.group_name] = my_orm

        return self.orm_class

    @classmethod
    def get_pom_classes(cls, session) -> Optional[List["PomSomMapper"]]:
        """
        Get the pom_classes from database data.

        Note that this method does not ensure that a ORM mapper is created
        for each class. Use pom_class.ensure_mapping() method to dynamically
        produce the ORM class

        :return: A list of SomPomMappers object for each class in the db
        """

        stmt = select(cls)
        pom_classes = session.execute(stmt).scalars().all()
        pom_class: PomSomMapper
        for pom_class in pom_classes:
            cls.pom_classes[pom_class.id] = pom_class

        return cls.pom_classes.values()

    @classmethod
    def get_pom_class_ids(cls, session):
        """
        Return all the pom_som_class ids as a list
        :return:
        """

        # stmt = select(cls.id)
        # pom_class_ids: Optional[List[str]] = session.execute(
        #     stmt).scalars().all()
        if cls.pom_classes and len(cls.pom_classes) > 0:
            pass
        else:
            cls.get_pom_classes(session)
        return list(cls.pom_classes.keys())

    @classmethod
    def get_pom_class(cls, pom_class_id: String, session):
        """
        Return the pom_class object for a given pom_class_id.

        See also Entity.get_orm_for_pom_class

        :param pom_class_id: the id of a pom_class
        """
        return session.get(cls, pom_class_id)

    @classmethod
    def get_pom_class_from_group(cls, group: KGroup, session=None):
        """Returns the PomSomMapper for a given group"""
        # TODO use property instead
        pom_id = getattr(group, "_pom_class_id", None)
        if pom_id is None:
            kname = group.kname
            pom_id = cls.get_pom_id_by_group_name(session, kname)
        if pom_id:
            return cls.get_pom_class(pom_id, session)
        return None

    @classmethod
    def get_pom_id_by_group_name(cls, session, kname):
        """ Returns the PomSomMapper id for a given group name"""
        for pom in cls.get_pom_classes(session):
            if kname == pom.group_name:
                # TODO use a setter
                return pom.id
        return None

    @classmethod
    def get_orm_for_group(cls, groupname: str):
        """
        PomSomMapper.get_orm_for_groupname("act")

        will return the ORM class corresponding to the groupname "act"
        """
        return PomSomMapper.group_orm_models.get(groupname, None)

    @classmethod
    def ensure_all_mappings(cls, session):
        """
        Ensures that every class currently defined in the database has
        a python table object and a python ORM object

        :return: None
        """

        pom_classes = cls.get_pom_classes(session=session)
        for pom_class in pom_classes:
            pom_class.ensure_mapping(session=session)

    def element_class_to_column(self, eclass: str, session: None) -> str:
        """ Return the column name corresponding to a group element class.

        Args:
            eclass (str): The class of an element (included in the export file).

        Returns:
            str: The name of the column corresponding to this element in the mapped table.
        """

        cattr: PomClassAttributes = None
        for cattr in self.class_attributes:
            if cattr.colclass == eclass:
                return cattr.colname
        if self.id != "entity":
            super_class = self.get_pom_class(self.super_class, session=session)
            if super_class is not None:
                return super_class.element_class_to_column(eclass, session=session)
        return None

    def column_to_class_attribute(
        self, colname: str, session: None
    ) -> "PomClassAttributes":
        """Return the class attribute corresponding to a column name.

        Args:
            colname (str): The name of a column in the mapped table.

        """
        for cattr in self.class_attributes:
            if cattr.colname == colname:
                return cattr
        if self.id != "entity":
            super_class = self.get_pom_class(self.super_class, session)
            if super_class is not None:
                return super_class.column_to_class_attribute(colname, session)
        return None

    @classmethod
    def kgroup_to_entity(cls, group: KGroup, session=None, with_pom=None) -> Entity:
        """
        Store a Kleio Group in the database.

        :param group: a Kleio Group
        :param with_pom: id of a PomSom class to handle storing this group
        :return: An ORM object that can store this group in the database

        To produce the ORM-POM representation of a group we need to find
        the PomSomMapper specific to that group, using the following:

            * if with_pom is given with a PomSomMapper id we fetch that
            * if the group was imported it should have _pom_class_id
            * if neither then we search for a PomSom mapper with "groupname"
               equal to the name of the group.
        """
        if with_pom is not None:
            pom_class_id = with_pom
            pom_class = cls.get_pom_class(pom_class_id, session)
        else:
            pom_class = cls.get_pom_class_from_group(group, session)

        if pom_class is None:
            raise ValueError(
                f"Could not determine PomSomMapper for this group: {group}"
            )

        pom_class.ensure_mapping(session)
        ormClass = Entity.get_orm_for_pom_class(pom_class.id)

        entity_from_group: Entity = ormClass()
        entity_from_group.groupname = group.kname

        # extra_info =  this will store the extra information in comment and original words
        extra_info: dict = dict()  # {el1:{'core':'','comment':'','original':''},el2:...}
        group_obs = ''  # this will store the observation information with extra info
        columns = inspect(ormClass).columns
        # temp
        for column in set([c.name for c in columns]):
            cattr = pom_class.column_to_class_attribute(column, session)
            if cattr is None:  # cols as updated and indexed not mapped
                continue
            if cattr.colclass == "id":
                pass
            element: KElement = group.get_element_by_name_or_class(cattr.colclass)
            if element is not None and element.core is not None:
                try:
                    if len(element.core) > cattr.colsize:
                        warnings.warn(
                            f"""Element {element.name} of group {group.kname}:{group.id}"""
                            f""" is too long for column {cattr.colname}"""
                            f""" of class {pom_class.id}"""
                            f""" truncating to {cattr.colsize} characters""",
                            stacklevel=2,
                        )
                        core_value = element.core[: cattr.colsize]

                    else:
                        core_value = element.core
                    setattr(entity_from_group, cattr.colname, str(core_value))
                    extra_info.update({element.name: {}})
                    if cattr.colname == "obs":
                        group_obs = core_value  # we save the obs element for later
                    if element.comment is not None and element.comment.strip() != '':
                        extra_info[element.name].update({'comment': element.comment.strip()})
                    if element.original is not None and element.original.strip() != '':
                        extra_info[element.name].update({'original': element.original.strip()})
                except Exception as e:
                    session.rollback()
                    raise ValueError(
                        f"""Error while setting column {cattr.colname}"""
                        f""" of class {pom_class.id} """
                        f"""with element {element.name}"""
                        f""" of group {group.kname}:{group.id}: {e} """
                    ) from e

        # positional information in the original file
        entity_from_group.the_source = getattr(group, 'source_id', None)  # set by import
        entity_from_group.the_line = group.line
        entity_from_group.the_level = group.level
        entity_from_group.the_order = group.order
        extra_info = {k: v for k, v in extra_info.items() if v != {}}
        # for elname, extras in extra_info.items():
        #     el_obs = ''
        #     if extras.get('comment', None) is not None and extras['comment'].strip() != '':
        #         el_obs = el_obs + f"# {extras['comment']}. "
        #     if extras.get('original', None) is not None and extras['original'].strip() != '':
        #         el_obs = el_obs + f"% {extras['original']}. "
        #     if el_obs != '':
        #         group_obs = f"{group_obs} {elname} {el_obs}"
        # if group_obs.strip() != '':
        #     entity_from_group.obs = group_obs.strip()
        if len(extra_info) > 0:
            if hasattr(entity_from_group, 'extra_info'):  # has JSON field
                entity_from_group.extra_info = extra_info
            else:  # no JSON field
                if group_obs is not None and group_obs.strip() != '':
                    group_obs = f"{group_obs.strip()} extra_info: {json.dumps(extra_info)}"
                else:
                    group_obs = f"extra_info: {json.dumps(extra_info)}"

        entity_from_group.obs = group_obs

        # check if we have an id from the group
        # check if this group is enclosed in another
        container_id = group.get_container_id()
        if container_id not in ["root", "None", "", None]:
            entity_from_group.inside = container_id
        else:
            entity_from_group.inside = None
        return entity_from_group

    @classmethod
    def store_KGroup(cls, group: KGroup, session=None):
        """Store a Kleio Group in the database

        Will recursively store all the groups included in this group.

        If the group is already in the database it will be deleted, as well
        as all included groups.

        This is the main method that import Kleio transcripts into the database.
        """
        if session is None:
            raise ValueError("No session provided")

        entity_from_group: Entity = cls.kgroup_to_entity(group, session)
        exists = session.get(entity_from_group.__class__, entity_from_group.id)

        if exists is not None:
            session.delete(exists)
            session.commit()

        try:
            session.add(entity_from_group)
            session.commit()
        except Exception as e:  # pylint: disable=broad-except
            raise e

        in_group: KGroup
        for in_group in group.contains():
            cls.store_KGroup(in_group, session)

        try:
            session.commit()
        except Exception as e:  # pylint: disable=broad-except
            raise e

    def __repr__(self):
        return (
            f'PomSomMapper(id="{self.id}", '
            f'table_name="{self.table_name}", '
            f'class_group="{self.group_name}", '
            f'super_class="{self.super_class}" '
            f")"
        )

    def __str__(self):
        r = f"{self.id} table {self.table_name} super {self.super_class}\n"
        for cattr in self.class_attributes:  # pylint: disable=no-member
            r = r + f"{cattr.the_class}.{cattr.name} \t"
            r = r + f"class {cattr.colclass} \t"
            r = r + f"col {cattr.colname} \ttype {cattr.coltype} "
            r = r + f"size {cattr.colsize} precision {cattr.colprecision}"
            r = r + f"primary key {cattr.pkey} \n"
        return r


class PomClassAttributes(Base):   # pylint: disable=too-few-public-methods
    """
    Attribute of a PomClass. Maps kleio group element to table columns

    the_class: id of the PomSomClass this attribute is attached to.
    name     : name of of the attribute
    colname  : name of the column in the corresponding table
    colclass : class of the column (element source if different from colname)
    coltype  : type of the column
    colsize  : size of the column, int
    colprecision: precision of the column (if decimal), int
    pkey     : if > 0 order of this column in the primary key of the table

    """

    __tablename__ = "class_attributes"

    the_class = Column(String, ForeignKey("classes.id"), primary_key=True)

    pom_class = relationship(
        "PomSomMapper", foreign_keys=[the_class], backref="class_attributes"
    )
    name = Column(String(32), primary_key=True)
    colname = Column(String(32))
    colclass = Column(String(32))
    coltype = Column(String)
    colsize = Column(Integer)
    colprecision = Column(Integer)
    pkey = Column(Integer)

    def __repr__(self):
        return (
            f'PomClassAttributes(the_class="{self.the_class}", '
            f'name="{self.name}", '
            f'colname="{self.colname}", '
            f'colclass="{self.colclass}", '
            f'coltype="{self.coltype}", '
            f'colsize="{self.colsize}", '
            f'colprecision="{self.colprecision}", '
            f'pkey="{self.pkey}" '
            f")"
        )



import random
import pytest
import tempfile

# Import the methods from __init__.py
from timelink.migrations import set_db_url, current, upgrade, downgrade
from timelink.migrations import ALEMBIC_CFG
from timelink.cli import parse_db_url, create_db_index
from timelink.api.database import TimelinkDatabase
from tests import reference_db_con_str, skip_on_travis


# Skip tests if running on Travis CI
pytestmark = skip_on_travis


@pytest.fixture
def timelink_db():
    temp_dir = tempfile.TemporaryDirectory()
    db_url = f"sqlite:///{temp_dir.name}/test.db"

    db = TimelinkDatabase(db_url=db_url)
    yield db


def test_create_reference_db():
    db: TimelinkDatabase = TimelinkDatabase(db_url=reference_db_con_str)
    print(db.db_table_names())
    result = upgrade(reference_db_con_str, revision="head")
    # Add assertions to check the database state if needed
    assert result is None


def test_set_db_url(timelink_db):
    db_url = timelink_db.db_url
    set_db_url(db_url)
    assert ALEMBIC_CFG.get_main_option("sqlalchemy.url") == db_url


def test_current(timelink_db):
    db_url = timelink_db.db_url
    set_db_url(db_url)
    current(db_url, verbose=True)
    # Add assertions to check the output if needed


def test_upgrade(timelink_db):
    db_url = timelink_db.db_url
    set_db_url(db_url)
    upgrade(db_url, revision="head")
    # Add assertions to check the database state if needed


def test_upgrade_random():
    # get list of databases avoiding _users table
    db_list = create_db_index(avoid_patterns=["_users"])
    for _ in range(5):   # try 5 random databases some are not upgradable
        db_key = list(db_list.keys())[random.randint(0, len(db_list) - 1)]
        db_url = parse_db_url(db_key)
        db: TimelinkDatabase = TimelinkDatabase(db_url=db_url)
        print(db.db_table_names())
        result = upgrade(db_url, revision="head")
        if result is None:
            break
        else:
            print(f"Failed to upgrade {db_url} {result}")
    # Add assertions to check the database state if needed
    assert result is None


def test_downgrade(timelink_db):
    db_url = timelink_db.db_url
    set_db_url(db_url)
    result = downgrade(db_url, revision="48dd")
    # Add assertions to check the database state if needed
    assert result is None



"""
.. module:: groups
   :synopsis: Classes for handling Kleio groups and elements.

.. moduleauthor: Joaquim Ramos de Carvalho

Kleio Groups are the building blocks for transcription of historical sources.

The classes in this package can be used to generate Kleio sources from data
fetched from databases or read from csv or json files.

The base classes are:

KKleio
    A Kleio document. It should include a single KSource group.

KSource
    A Historical Source. Can contain a variable number
    of KAct groups.

KAct
    A historical act. Can contain a variable number
    of KPerson or KObject.

In normal usage the basic groups are extended for a particular context:



.. code-block:: python


        from timelink.kleio.groups import KKleio, KSource, KAct,
        KAbstraction, KPerson, KLs, KAtr


        kleio = KKleio

        fonte = KSource.extend('fonte',
                                also=['tipo', 'data', 'ano', 'substitui',
                                        'loc', 'ref', 'obs'])

        lista = KAct.extend('lista',
                            position=['id', 'dia', 'mes', 'ano'],
                            guaranteed=['id', 'ano', 'mes', 'dia'],
                            also=['data', 'tipo', 'loc', 'obs'])

        auc = KAbstraction.extend('auc', position=['name', ''], also=[
            'level', 'id'], guaranteed=['id'])

        n = KPerson.extend('n', position=['nome', 'sexo'], guaranteed=[
            'id', 'nome', 'sexo'], also=['mesmo_que', 'obs'])

        pai = KPerson.extend('pai', position=['nome'], guaranteed=[
            'id', 'nome'], also=['mesmo_que', 'obs'])

        mae = KPerson.extend('mae', position=['nome'], guaranteed=[
            'id', 'nome'], also=['mesmo_que', 'obs'])

        ls = KLs.extend('ls',
                        position=['type', 'value', 'data'],
                        also=['data', 'obs'])

        atr = KAtr.extend(
            'atr', position=['type', 'value', 'data'], also=['data', 'obs'])

        n.allow_as_part(pai)

        n.allow_as_part(mae)



"""

from timelink.kleio.groups.kabstraction import KAbstraction  # noqa
from timelink.kleio.groups.kobject import KObject  # noqa
from timelink.kleio.groups.kgeoentity import KGeoentity  # noqa
from timelink.kleio.groups.kact import KAct  # noqa
from timelink.kleio.groups.kattribute import KAttribute   # noqa
from timelink.kleio.groups.kls import KLs  # noqa
from timelink.kleio.groups.katr import KAtr  # noqa
from timelink.kleio.groups.kelement import (KElement,  # noqa
                                            KDate,
                                            KDay,
                                            KMonth,
                                            KYear,
                                            KType,
                                            KValue,
                                            KId,
                                            KEntityInAttribute,
                                            KOriginInRel,
                                            KReplace,
                                            KSameAs,
                                            KXSameAs,
                                            KName,
                                            KSex,
                                            KDescription,
                                            KObs,
                                            KLevel,
                                            KLine,
                                            KOrder,
                                            KStructure,
                                            KLevel,
                                            KLoc,
                                            KRef,
                                            KDestName,
                                            KDestId,
                                            KSummary,
                                            KReplaceSourceId,
                                            )  # noqa
from timelink.kleio.groups.kgroup import KGroup  # noqa
from timelink.kleio.groups.kkleio import KKleio  # noqa
from timelink.kleio.groups.kobject import KObject  # noqa
from timelink.kleio.groups.kperson import KPerson  # noqa
from timelink.kleio.groups.krelation import KRelation  # noqa
from timelink.kleio.groups.ksource import KSource  # noqa



# flake8: noqa: B008

"""
FastAPI app for timelink

Next

* Authentication

    * startup Fief docker container, try to add first user
    * add first admin user to database

To Run

    * source .venv/bin/activate; cd timelink/app; uvicorn main:app --reload

To Test

* http://127.0.0.1:8008/docs
* http://127.0.0.1:8008/redoc

To debug

* https://fastapi.tiangolo.com/tutorial/debugging/

"""
# Standard library imports
import os
from datetime import date, timedelta
from enum import Enum
from typing import Annotated, List, Optional

# Third-party imports
from fastapi import FastAPI, Depends, Request, status
from fastapi import HTTPException, Query
from fastapi.responses import HTMLResponse
from fastapi.security import OAuth2PasswordBearer
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi.security import OAuth2PasswordBearer
from fastapi.security import OAuth2PasswordBearer
from fastapi.security import OAuth2PasswordRequestForm

from fastui import FastUI, AnyComponent, prebuilt_html, components as c
from fastui.components.display import DisplayMode, DisplayLookup
from fastui.events import GoToEvent, BackEvent

# import realted with starlette admin app
from starlette_admin import EnumField

from starlette_admin.contrib.sqla import Admin, ModelView
from starlette_admin.views import Link


# imports related to authentication with fief
from fastapi.security import OAuth2AuthorizationCodeBearer
from fief_client import FiefAccessTokenInfo, FiefAsync
from fief_client.integrations.fastapi import FiefAuth

from pydantic import BaseModel, Field

from jinja2 import Environment, PackageLoader
from sqlalchemy.orm import Session
import uvicorn

# Local application/library specific imports
from timelink import version
from timelink.api import crud, models, schemas
from timelink.api.database import TimelinkDatabase, is_valid_postgres_db_name
from timelink.api.schemas import EntityAttrRelSchema, ImportStats

from timelink.app.backend.settings import Settings
from timelink.app.backend.timelink_webapp import TimelinkWebApp
from timelink.app.dependencies import get_current_user, get_kleio_server
from timelink.app.models.user import User, UserProperty
from timelink.app.models.user_database import UserDatabase
from timelink.app.models.project import Project, ProjectAccess, AccessLevel
from timelink.kleio import (
    KleioServer,
    api_permissions_normal,  # todo: should have kleio prefixed name
    kleio_server as kserver,
    token_info_normal,  # todo: chenge to kleio prefixed names
)
from timelink.kleio.importer import import_from_xml
from timelink.kleio.kleio_server import KleioServer
from timelink.kleio.schemas import ApiPermissions, KleioFile, TokenInfo
from timelink.app.dependencies import get_current_active_user, get_db
from timelink.app.schemas.user import UserSchema
from timelink.app.services.auth import auth

from timelink.app.web import router as fastui_router

# Get Pydantic-based settings defined in timelink.app.backend.settings
settings = Settings(timelink_admin_pwd="admin")

# Move to settings?
ACCESS_TOKEN_EXPIRE_MINUTES = 30

app = FastAPI()
app.include_router(fastui_router, prefix="/fastui")


# create TimelinkWebApp instance
if hasattr(app.state, "webapp") is False or app.state.webapp is None:
    initial_users = []
    possible_timelink_home = KleioServer.find_local_kleio_home()
    # End deprecated

    # need to add timelink_home to settings and to webapp
    webapp = TimelinkWebApp(
        app_name=settings.timelink_app_name,
        timelink_home=possible_timelink_home,
        users_db_name=settings.timelink_users_db_name,
        users_db_type=settings.timelink_users_db_type,
        initial_users=initial_users,

    )

    app.state.webapp = webapp
    app.state.status = "Initialized"

admin = Admin(webapp.users_db.engine,
              title="timelink admin",
              )
admin.add_view(ModelView(User))
admin.add_view(ModelView(UserProperty))
admin.add_view(ModelView(ProjectAccess))
admin.add_view(ModelView(Project))
admin.add_view(Link(label="Timelink web", icon="fa fa-link", url="/"))

admin.mount_to(app)

app.mount("/static", StaticFiles(packages=[("timelink")]), name="static")

# this is how to load the templates from inside the package
env = Environment(loader=PackageLoader("timelink", "app/templates"))
templates = Jinja2Templates(env=env)


@app.get("/test_token", response_model=UserSchema)
async def test_token(
    current_user: Annotated[UserSchema, Depends(get_current_active_user)],
):
    return current_user


@app.get("/web/show/{id}", response_class=HTMLResponse)
async def show_item(request: Request, id: str):
    # problem here is that I need the current project
    return templates.TemplateResponse(
        request=request, name="item.html", context={"id": id}
    )


@app.post("/web/search/", response_model=List[schemas.SearchResults])
async def search(search_request: schemas.SearchRequest):
    """Search for items in the database.

    Args:
        search_request: SearchRequest object

    Returns:
        Search results
    """
    result1 = schemas.SearchResults(
        id="jrc",
        the_class="person",
        description="Joaquim Carvalho: " + repr(search_request),
        start_date=date(1958, 5, 24),
        end_date=date(2023, 1, 4),
    )
    result2 = schemas.SearchResults(
        id="mag",
        the_class="person",
        description="Magda Carvalho: " + repr(search_request),
        start_date=date(1960, 1, 1),
        end_date=date(2023, 1, 4),
    )
    if search_request.q == "jrc":
        return [result1]
    if search_request.q == "mag":
        return [result2]
    if search_request.q == "both":
        return [result1, result2]
    return []


@app.post("/syspar/", response_model=models.SysParSchema)
async def set_syspar(
    syspar: models.SysParSchema, db: Annotated[Session, Depends(get_db)]
):
    """Set system parameters

    Args:
        syspar: SysPar object

    Returns:
        SysPar object
    """
    return crud.set_syspar(db, syspar)


@app.get("/syspar/", response_model=list[models.SysParSchema])
async def get_syspars(
    q: list[str] | None = Query(
        default=None,
        title="Name of system parameter",
        description="Multiple values allowed," "if empty return all",
    ),
    db: Session = Depends(get_db),
):
    """Get system parameters

    Args:
        q: query string, multiple values allowed, if empty return all

    """

    return crud.get_syspar(db, q)


@app.post("/syslog", response_model=models.SysLogSchema)
async def set_syslog(syslog: models.SysLogCreateSchema, db: Session = Depends(get_db)):
    """Set a system log entry"""
    return crud.set_syslog(db, syslog)


@app.get("/syslog", response_model=list[models.SysLogSchema])
async def get_syslog(
    nlines: int | None = Query(
        default=10,
        title="Get last N lines of log",
        description="If number of lines not specified" "return last 10",
    ),
    db: Session = Depends(get_db),
):
    """Get log lines

    Args:
        nlines: number of most recent lines to return

    TODO: add fitler since (minutes, seconds)
    """
    result = crud.get_syslog(db, nlines)
    return result


@app.get("/sources/import-file/{file_path:path}", response_model=ImportStats)
async def import_file(file_path: str, db: Session = Depends(get_db)):
    """Import kleio data from xml file"""
    result = import_from_xml(file_path, db, {"return_stats": True, "mode": "TL"})
    response = ImportStats(**result)

    return response


@app.get("/get/{id}", response_model=EntityAttrRelSchema)
async def get(id: str, db: Session = Depends(get_db)):
    """Get entity by id"""
    return crud.get(db, id)


# Kleio server interface
@app.get("/kleio/is-running", response_model=bool)
async def is_kleio_server_running():
    """Check if kleio server is running"""
    return kserver.is_kserver_running()


# invalidate user
@app.get("/kleio/invalidate-user/{user}", response_model=str)
async def invalidate_user(
    user: str, kserver: KleioServer = Depends(get_kleio_server)
):  # noqa: B008
    """Invalidate a user"""
    return kserver.invalidate_user(user)


# generate token for user
@app.get("/kleio/generate-normal-token/{user}", response_model=str)
async def generate_norma_token(
    user: str, kserver: KleioServer = Depends(get_kleio_server)  # noqa: B008
):
    """Generate a token for a user"""
    return kserver.generate_token(user, token_info_normal)


# get translation status
@app.get("/kleio/translation-status/{path:path}", response_model=list[KleioFile])
async def translation_status(
    path: str,
    recurse: str,
    status: str,
    kserver: KleioServer = Depends(get_kleio_server),  # noqa: B008
):
    """Get translations from kleio server

    Args:
        path (str): path to the directory in sources.
        recurse (str): if "yes" recurse in subdirectories.
        status (str): filter by translation status:
                        V = valid translations;
                        T = need translation (source more recent than translation);
                        E = translation with errors;
                        W = translation with warnings;
                        P = translation being processed;
                        Q = file queued for translation;
                        * = all
    """
    if status == "*":
        status = None
    return kserver.get_translations(path, recurse, status)


# translate
@app.get("/kleio/translate/{path:path}", response_model=str)
async def translate(
    path: str,
    recurse: str,
    spawn: str,
    kserver: KleioServer = Depends(get_kleio_server),  # noqa: B008
):
    """Translate sources from kleio server

    Args:
        path (str): path to the file or directory in sources.
        recurse (str): if "yes" recurse in subdirectories.
        spawn (str): if "yes" spawn a translation process for each file.
    """
    return kserver.translate(path, recurse, spawn)


# Web zone

# web prefix uses templates for htmx and tailwind
@app.get("/web")
async def root(request: Request,
               user: UserSchema = Depends(get_current_user)
               ):
    """Timelink API end point. Check URL/docs for API documentation."""
    webapp: TimelinkWebApp = request.app.state.webapp
    context = {
        "welcome_message": "Welcome to Timelink API and Webapp",
        "webapp_info": webapp.get_info(),
        "projects": webapp.projects,
        "user": user,
    }
    return templates.TemplateResponse(
        request=request, name="index.html", context=context
    )


@app.get("/web/templates/{template_name}")
async def get_template(request: Request, template_name: str):
    """Get a template"""
    webapp: TimelinkWebApp = request.app.state.webapp
    context = {
        "welcome_message": "Welcome to Timelink API and Webapp",
        "webapp_info": webapp.get_info(),
        "webapp_projects": webapp.projects,
    }

    return templates.TemplateResponse(
        request=request, name=template_name, context=context
    )


# This is the main entry point for the fastUI interface
# Must come last as it matches all paths
@app.get("/{path:path}")
async def html_landing() -> HTMLResponse:
    """Simple HTML page which serves the React app, comes last as it matches all paths."""
    return HTMLResponse(prebuilt_html(title="Timelink", api_root_url="/fastui"))


# Tutorial
class ModelName(str, Enum):
    """Enum for model names"""

    alexnet = "alexnet"
    resnet = "resnet"
    lenet = "lenet"


fake_items_db = [{"item_name": "Foo"}, {"item_name": "Bar"}, {"item_name": "Baz"}]


@app.get("/models/{model_name}")
async def get_model(model_name: ModelName):
    """Example of path parameters with enum validation"""
    if model_name == ModelName.alexnet:
        return {"model_name": model_name, "message": "Deep Learning FTW!"}

    if model_name.value == "lenet":
        return {"model_name": model_name, "message": "LeCNN all the images"}

    return {"model_name": model_name, "message": "Have some residuals"}


@app.get("/files/{file_path:path}")
async def read_file(file_path: str):
    """Example of path parameters"""
    return {"file_path": file_path}


@app.get("/items/")
async def read_item(skip: int = 0, limit: int = 10):
    """Example of query parameters"""
    return fake_items_db[skip : skip + limit]


@app.get("/items2/{item_id}")
async def read_item2(item_id: str, q: str | None = None):
    """Example of query parameters with optional
    http://127.0.0.1:8000/items2/foo?q=somequery
    http://127.0.0.1:8000/items2/foo
    """
    if q:
        return {"item_id": item_id, "q": q}
    return {"item_id": item_id}


# Sandbox
# testing specifying the database in the path
@app.get("/{dbname}/id/{id}", response_model=dict)
async def get_id(eid: str, dbname: str, db: Session = Depends(get_db)):  # noqa: B008
    """get entity with id from database

    This will pass the name of the database in the path
    to the get_db() function
    """
    result = {
        "database": dbname,
        "id": f"info for id {eid}",
        "url": repr(db.get_bind().url),
    }
    return result


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8008)
    app.state.status = "Running"



import os
import yaml
import json
from pprint import pprint


json_path = "/Users/jrc/develop/timelink-py/tests/timelink-home/projects/test-project/structures/sources.str.json"

with open(json_path) as stream:
    pystr = json.load(stream)
    groups = pystr['groups']

# print current directory
print(os.getcwd())

path = 'tests/timelink-home/projects/test-project/structures/yaml/sample-stru.yml'
with open(path) as stream:
    pyaml = yaml.full_load(stream)

pprint(pyaml)



# flake8: noqa:F401
""" Handling of Database persistence for the Person Oriented Model (POM).

The classes on this module represent the core entities in the POM: entities,sources,
acts, people, objects, attributes and relations. SQLAlchemy is used as the
underlying ORM library.

They allow storage and query and also dynamic mapping of new types of
entities defined in ``kleio`` source files, through the PomSomMapper class.

For a detailed description of *Timelink* data models see: :doc:`som_pom_mapping`.

Also check  :class:`timelink.mhk.models.pom_som_mapper.PomSomMapper`

(c) Joaquim Carvalho 2021.
MIT License, no warranties.

"""


from .base_class import Base  # noqa pylint: disable=unused-import
from .entity import Entity  # noqa pylint: disable=unused-import
from .attribute import Attribute  # noqa pylint: disable=unused-import
from .relation import Relation  # noqa pylint: disable=unused-import
from .act import Act  # noqa pylint: disable=unused-import
from .source import Source  # noqa pylint: disable=unused-import
from .person import Person  # noqa pylint: disable=unused-import
from .object import Object  # noqa pylint: disable=unused-import
from .pom_som_mapper import PomSomMapper, PomClassAttributes  # noqa pylint: disable=unused-import
from .base_mappings import pom_som_base_mappings  # noqa pylint: disable=unused-import
from .system import (SysPar, SysParSchema, 
                         SysLog, SysLogSchema, SysLogCreateSchema)  # noqa pylint: disable=unused-import



""" Utilities for styling pandas dataframes"""
from matplotlib import colors
import matplotlib.pyplot as plt
import numpy as np


def category_palette(categories, cmap_name=None):
    """Create a color palette associated with a list of categories

    Args:
        Categories: List of categories
        cmap_name: matplotlib color map defaults to Pastel2
            see https://matplotlib.org/stable/tutorials/colors/colormaps.html

    Returns a dict with the categories as keys and colors as values
    """
    if cmap_name is None:
        cmap_name = "tab20"

    # Assuming 'cmap_name' is the name of your colormap and 'categories' is a list of categories
    num_categories = len(categories)
    cmap = plt.colormaps[cmap_name]
    palette = cmap(np.linspace(0, 1, num_categories))

    mapcolors = [colors.rgb2hex(palette[i]) for i in range(num_categories)]
    # shufle the colors
    # random.shuffle(mapcolors)

    cat_to_color = dict(zip(categories, mapcolors))
    return cat_to_color


def style_color_row_by_category(row, palette, category="id"):
    """Color row by category. Function for styling dataframes
    Usage: display(df.style.apply(style_color_row_by_category,axis=1,palette=mypalette))

    Args
        row: this is passed by pandas when rendering the dataframe
        palette: a dict that maps category values to colors
        category:  column that determines the row color
    """
    id = row[category]
    row_colors = [f"background-color: {palette.get(id,'#ffffff')}"] * len(row)
    return row_colors


def styler_row_colors(df, category="id", columns=None, palette=None, cmap_name=None):
    """returns a dataframe setting the row color according to a category

    Args:
        df: dataframe
        category: name of column with category that determines color, defaults to id
        cmap_name; name of matplolib color map to use, defaults to 'Pastel2'

    """
    categories = df[category].unique()
    if cmap_name is None:
        cmap_name = "Pastel2"
    if palette is None:
        palette = category_palette(categories, cmap_name=cmap_name)
    if columns is None:
        columns = df.columns
    if category not in columns:
        cols = columns + [category]
    else:
        cols = columns

    r = df[cols].style.apply(
        style_color_row_by_category, axis=1, category=category, palette=palette
    )
    return r



""" Interface to Kleio Server"""

import logging
import re
import socket
import os
from time import sleep
import time
from typing import List, Optional, Tuple
import docker
import secrets
import requests
from jsonrpcclient import request, Error, Ok, parse
import urllib.request

from .schemas import KleioFile, TokenInfo


class KleioServerException(Exception):
    pass


class KleioServerForbidenException(KleioServerException):
    """Corresponds to json_rpc error code -32006 or HTTP 403"""

    pass


class KleioServerDockerException(KleioServerException):
    """Problem with docker"""

    pass


class KleioServer:

    """This class interfaces to a Kleio server through its JSON-RPC api.
    It also provides convenience methods
    to start a server in Docker locally.

    This class is not intended to be instantiated directly.
    Use KleioServer.start() and KleioServer.attach() to create instances of KleioServer.

    Args:
        container (docker.models.containers.Container): runing kleio server container
        url (str): kleio server url if running in a different machine (container=None)
        token (str): kleio server token if running in a different machine (container=None)
        kleio_home (str): kleio server home directory.
                          If None and container is not None
                            then kleio_home is obtained
                            from the container. If not none


    Main methods:

        - start: Starts a kleio server in docker.
        - attach: Attach to an already running Kleio Server.
        - get_server: Return a running kleio server associated with a kleio_hom
        - is_server_running: Check if a kleio server is running in docker mapped to a given kleio home directory.
        - find_local_kleio_home: Find kleio home directory.
        - make_token: Get the kleio server token from environment or generate a new one if not set.
        - get_token: Get the running kleio server token.
        - get_kleio_home: Get the kleio server home directory.
        - get_container: Get the kleio server container.
        - get_logs: Get the logs of the kleio server container.
        - get_url: Get the kleio server url.
        - call: Basic call to kleio server API.
        - stop: Stop the kleio server container.
        - invalidate_user: Invalidate a user.
        - generate_token: Generate a token for a user.
        - get_translations: Get translation status from kleio server.
        - translate: Translate sources from kleio server.
        - translation_clean: Clean translations from kleio server.
        - get_sources: Get sources from kleio server.
        - get_report: Get report from kleio server.
        - get_url_content: Get content from Kleio Server.
        - get_home_page: Get home page from Kleio Server.
    """

    #: kleio server host
    host: str
    #: kleio server url
    url: str
    #: kleio server admin token
    kleio_admin_token: str
    #: kleio server home directory
    kleio_home: str
    #: kleio server container
    container: docker.models.containers.Container

    @staticmethod
    def start(
        kleio_image: str = "timelinkserver/kleio-server",
        kleio_version: str | None = "latest",
        kleio_home: str | None = None,
        kleio_admin_token: str | None = None,
        kleio_server_port="8088",
        kleio_external_port=None,
        kleio_server_workers="3",
        kleio_idle_timeout=900,
        kleio_conf_dir=None,
        kleio_source_dir=None,
        kleio_stru_dir=None,
        kleio_token_db=None,
        kleio_default_stru=None,
        kleio_debug=None,
        consistency: str = "cached",
        port: int = None,
        update: bool = False,
        reuse: bool = True,
        stop_duplicates: bool = False,
    ):
        """Starts a kleio server in docker

        Args:
            kleio_image (str): kleio server image, defaults to "timelinkserver/kleio-server"
            kleio_version (str, optional): kleio-server image version, defaults to "latest"
            kleio_home (str, optional): kleio home directory,
                                        defaults to None -> current directory
            kleio_token (str, optional): kleio server admin token,
                                    defaults to None -> generate a random token
            kleio_server_port (str, optional): kleio server port (in the container),
                                                 defaults to "8088".
            kleio_external_port (str, optional): kleio server external port (in the host),
                                                    defaults to "8089".
            kleio_server_workers (str, optional): number kleio server workers, defaults to "3".
            kleio_idle_timeout (int, optional): kleio server idle timeout, defaults to 900.
            kleio_conf_dir (str, optional): kleio server configuration directory,
                                            defaults to None.
            kleio_source_dir (str, optional): kleio server sources directory,
                                            defaults to None.
            kleio_stru_dir (str, optional): kleio server structures directory,
                                            defaults to None.
            kleio_token_db (str, optional): kleio server token database,
                                            defaults to None.
            kleio_default_stru (str, optional): kleio server default structure,
                                            defaults to None.
            kleio_debug (str, optional): kleio server debug level, defaults to None.
            consistency (str, optional): consistency of the volume mount, defaults to "cached"
            port (int, optional): port to map to 8088,
                                    defaults to None -> find a free port starting at 8088
            update (bool, optional): update kleio server image, defaults to False
            reuse (bool, optional): if True, reuse an existing kleio server container
                                    with same keio_home, defaults to True.
            stop_duplicates (bool, optional): if True, stop and remove duplicate containers
                                    with same kleio_home, defaults to False.

        Returns:
            KleioServer: KleioServer object
        """
        if kleio_image is None:
            kleio_image = "timelinkserver/kleio-server"
        if kleio_version is None:
            kleio_version = "latest"
        if not is_docker_running():
            raise RuntimeError("Docker is not running")
        # TODO: test if kleio_home is a valid directory

        container: docker.models.containers.Container = start_kleio_server(
            image=kleio_image,
            version=kleio_version,
            kleio_home=kleio_home,
            kleio_admin_token=kleio_admin_token,
            kleio_server_port=kleio_server_port,
            kleio_external_port=kleio_external_port,
            kleio_server_workers=kleio_server_workers,
            kleio_idle_timeout=kleio_idle_timeout,
            kleio_conf_dir=kleio_conf_dir,
            kleio_source_dir=kleio_source_dir,
            kleio_stru_dir=kleio_stru_dir,
            kleio_token_db=kleio_token_db,
            kleio_default_stru=kleio_default_stru,
            kleio_debug=kleio_debug,
            consistency=consistency,
            update=update,
            reuse=reuse,
            stop_duplicates=stop_duplicates,
        )

        kserver = KleioServer(container)

        # wait for the server to be ready
        start_time = time.time()
        stop_time = 1
        timeout = 15
        while True:
            try:
                kserver.get_home_page()
                break
            except Exception as e:
                elapsed_time = time.time() - start_time
                if elapsed_time >= timeout:
                    raise RuntimeError(
                        f"Failed to start Kleio server after {timeout} seconds: {e}"
                    )
                logging.warning(
                    f"Kleio server not ready, retrying in {stop_time} seconds: {e}"
                )
                time.sleep(stop_time)
        logging.info("Kleio server started successfully")

        return kserver

    @staticmethod
    def attach(url: str, token: str, kleio_home: str = "."):
        """Attach to a already running Kleio Server.

        Use this either to attach to a running server
        outside docker (for instance in a Prolog session)
        or to a server running in another machine (should use
        htpps url in that case)

        Args:
            url (str): kleio server url
            token (str): kleio server token
            kleio_home (str, optional): kleio server home directory

        Returns:
            KleioServer: KleioServer object

        TODO #33:
        from urllib.parse import urlparse

        url = 'http://user:token@localhost:8088/path_to_kleio_home'
        parsed_url = urlparse(url)

        print('scheme:', parsed_url.scheme)
        print('netloc:', parsed_url.netloc)
        print('path:', parsed_url.path)
        print('username:', parsed_url.username)
        print('password:', parsed_url.password)
        print('hostname:', parsed_url.hostname)
        print('port:', parsed_url.port)

        """
        return KleioServer(url=url, token=token, kleio_home=kleio_home)

    @staticmethod
    def get_server(kleio_home: str = None, kleio_version="latest"):
        """Check if a kleio server is running in docker mapped to a given kleio home directory.

        If yes return a KleioServer object, otherwise return None
        If a specific version is required, it can be specified in kleio_version

        Args:
            kleio_home (str, optional): kleio home directory;
                                        defaults to None -> any kleio home.

        Returns:
            KleioServer or None: KleioServer object or None
        """
        if is_docker_running() is False:
            raise Exception("Docker is not running")

        container = get_kserver_container(
            kleio_home=kleio_home, kleio_version=kleio_version
        )
        if container is not None:
            return KleioServer(container)
        else:
            return None

    @staticmethod
    def is_server_running(kleio_home: str = None, kleio_version="latest"):
        """Check if a kleio server is running in docker mapped to a given kleio home directory.

        Args:
            kleio_home (str, optional): kleio home directory;
                                        defaults to None -> any kleio home.
            kleio_version (str, optional): kleio server version; defaults to "latest"
                                            if not specified only checks latest
        Return True of False

        Args:
            kleio_home (str, optional): kleio home directory;
                                        defaults to None -> any kleio home.

        Returns:
            bool: True if server is running, False otherwise
        """
        if is_docker_running() is False:
            raise Exception("Docker is not running")

        container = get_kserver_container(
            kleio_home=kleio_home, kleio_version=kleio_version
        )
        return container is not None

    @staticmethod
    def find_local_kleio_home(path: str = None):
        """Find kleio home directory.

        Kleio home directory is the directory where Kleio Server finds sources and auxiliary
        files like structures, mappings and inferences.

        It can be in the current directory, parent directory, or tests directory.
        It can be named "kleio-home", "timelink-home", or "mhk-home".

        A special case is when the current directory is "notebooks".
        In this case, kleio-home is assumed to be the parent directory of "notebooks"
        and thus set up as the timelink-project template.

        Args:
            path (str, optional): path to start searching from;
                                defaults to None -> current directory.

        Returns:
            str: kleio home directory
        """
        return find_local_kleio_home(path=path)

    @staticmethod
    def make_token():
        """Get the kleio server token from environment or generate a new one if not set.

        Returns:
            str: kleio server token
        """
        token = os.environ.get("KLEIO_ADMIN_TOKEN")
        if token is None:
            token = random_token()
            os.environ["KLEIO_ADMIN_TOKEN"] = token
        return token

    def __init__(
        self,
        container: docker.models.containers.Container = None,
        url: str = None,
        token: str = None,
        kleio_home: str = None,
    ):
        """Not to be used directly.

        See :py:meth:`KleioServer.start` and :meth:`KleioServer.attach`

        To start a kleio server locally in docker use :meth:`KleioServer.start`

        To attach to a running kleio server, local or remote, use :meth:`KleioServer.attach`

        Args:
            container (docker.models.containers.Container): runing kleio server container;
                                        if container is None, then url, token and kleio_home
                                        must be provided.
            url (str): kleio server url if running in a different machine (container=None)
            token (str): kleio server token if running in a different machine (container=None)
            kleio_home (str): kleio server home directory. If None and container is not None
                                    then kleio_home is obtained from the container.
                                    If not none
        """
        if container is None:
            if url is None:
                raise ValueError("url must be provided if container is None")
            if token is None:
                raise ValueError("token must be provided if container is None")
            if kleio_home is None:
                raise ValueError("kleio_home must be provided if container is None")
            self.url = url

            self.kleio_admin_token = token
            self.kleio_home = kleio_home
            self.container = None
            return

        if is_docker_running() is False:
            raise Exception("Docker is not running")

        self.container = container
        port = None
        try:
            p = list(container.attrs["HostConfig"]["PortBindings"])[0]
            port = container.attrs["HostConfig"]["PortBindings"][p][0]["HostPort"]
        except Exception as e:
            error_message = f"Could not get port from container {e}"
            logging.error(error_message)
            raise KleioServerDockerException(error_message)

        self.url = f"http://127.0.0.1:{port}"
        self.kleio_home = [
            mount["Source"]
            for mount in container.attrs["Mounts"]
            if mount["Destination"] == "/kleio-home"
        ][0]
        self.kleio_admin_token = [
            env
            for env in container.attrs["Config"]["Env"]
            if env.startswith("KLEIO_ADMIN_TOKEN")
        ][0].split("=")[1]

    def get_token(self):
        """Get the kleio server token

        Returns:
            str: kleio server token
        """
        return self.kleio_admin_token

    def get_kleio_home(self):
        """Get the kleio server home directory

        Returns:
            str: kleio server home directory
        """
        return self.kleio_home

    def get_container(self):
        """Get the kleio server container

        Returns:
            docker.models.containers.Container: kleio server container
        """
        return self.container

    def get_logs(self, **kwargs):
        """Get the logs of the kleio server container

        Args:
            **kwargs: arguments to pass to docker container logs:
            tail (str): number of lines to show from the end of the logs
            since (str): show logs since a datetime timestamp integer seconds
            until (str): show logs until a datetime timestamp integer seconds
            timestamps (bool): show timestamps
            follow (bool): follow log output

        Returns:
            str: kleio server logs

        """

        return self.container.logs(**kwargs).decode("utf-8")

    def get_url(self):
        """Get the kleio server url

        Returns:
            str: kleio server url
        """
        return self.url

    def __str__(self):
        return f"KleioServer(url={self.url}, kleio_home={self.kleio_home})"

    def call(self, method: str, params: dict, token: str = None, timeout=60):
        """Call kleio server API

        Args:
            method (str): kleio server API method
            params (dict): kleio server API method parameters
            token (str, optional): kleio server token; defaults to None -> use admin token
            timeout (int, optional): timeout in seconds; defaults to 60.

        Returns:
            dict: kleio server API response
        """
        url = f"{self.url}/json/"

        if token is None:
            token = self.kleio_admin_token

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
        }
        # we add the token to the params
        params["token"] = token
        rpc = request(method, params=params)
        response = requests.post(url, json=rpc, timeout=timeout, headers=headers)
        parsed = parse(response.json())
        if isinstance(parsed, Ok):
            return parsed.result
        elif isinstance(parsed, Error):
            code, message, data, id = parsed
            msg = f"Error {code}: {message} ({data} id:{id})"
            if code == -32006:
                raise KleioServerForbidenException("Forbiden " + msg)
            else:
                raise KleioServerException(f"Error {code}: {message} ({data} id:{id})")
        return response

    def stop(self):
        """Stop the kleio server container"""
        logging.debug(f"Stopping kleio server {self.container.name}")
        self.container.reload()
        if self.container.status != "running":
            self.container.reload()
            if self.container.status != "running":
                sleep(5)
            self.container.reload()
            if self.container.status != "running":
                raise KleioServerDockerException("Container is not running")
        self.container.stop()
        try:
            self.container.remove()
        except Exception as e:
            print(f"Error removing container: {e} {self.container.name}")
            pass
        return

    def invalidate_user(self, user: str):
        """Invalidate a user

        Args:
            user (str): user to invalidate

        Returns:
            dict: kleio server API response
        """
        pars = {"user": user}
        return self.call("users_invalidate", pars)

    def generate_token(self, user: str, info: TokenInfo):
        """Generate a token for a user

        Args:
            user (str): user to generate token for
            info (TokenInfo): token information

        Returns:
            dict: kleio server API response
        """
        pars = {"user": user, "info": info.model_dump()}
        return self.call("tokens_generate", pars)

    def get_translations(
        self, path: str, recurse: str = True, status: str = None, token: str = None
    ) -> List[KleioFile]:
        """Get translation status from kleio server.

        Returns the translation status of the kleio files in path.

        Args:
            path (str): Path to the directory in sources.
            recurse (str): If "yes", recurse in subdirectories.
            status (str, optional): Filter by translation status. Options include:

                V = valid translations
                T = need translation (source more recent than translation)
                E = translation with errors
                W = translation with warnings
                P = translation being processed
                Q = file queued for translation

            token (str, optional): Kleio server token.

        Returns:
            list[KleioFile]: List of KleioFile objects.

        """
        if recurse is True:
            recurse = "yes"
        elif recurse is False:
            recurse = "no"

        if path is None:
            path = ""

        path = str(path)

        if status is None:
            pars = {"path": path, "recurse": recurse}
        else:
            pars = {"path": path, "recurse": recurse, "status": status}
        translations = self.call("translations_get", pars, token=token)
        result = []
        for t in translations:
            kfile = KleioFile(**t)
            result.append(kfile)
        return result

    def translate(
        self, path: str, recurse: str = "yes", spawn: str = "yes", token=None
    ):
        """Translate sources from kleio server

        :param path: path to the directory in sources
        :type path: str
        :param recurse: if "yes" recurse in subdirectories
        :type recurse: str, optional
        :param spawn: if "yes" spawn a translation process for each file
        :type spawn: str, optional

        :return: kleio server API response
        :rtype: dict
        """
        pars = {"path": path}
        if recurse is True:
            recurse = "yes"
        elif recurse is False:
            recurse = "no"

        if recurse is not None:
            pars["recurse"] = recurse
        if spawn is not None:
            pars["spawn"] = spawn

        return self.call("translations_translate", pars), token

    def translation_clean(self, path: str, recurse: str):
        """clean translations from kleio server

        Removes translation results from kleio server.

        :param path: path to the directory in sources
        :type path: str
        :param recurse: if "yes" recurse in subdirectories
        :type recurse: str

        :return: kleio server API response
        :rtype: dict
        """
        pars = {"path": path, "recurse": recurse}
        return self.call("translations_delete", pars)

    def get_sources(self, path: str, recurse: bool = True, token=None):
        """Get list of sources from kleio server

        :param path: path to the directory in sources
        :type path: str
        :param recurse: if "yes" recurse in subdirectories
        :type recurse: str

        :return: kleio server API response
        :rtype: dict
        """
        if recurse is True:
            recurse = "yes"
        elif recurse is False:
            recurse = "no"

        pars = {"path": path, "recurse": recurse}
        return self.call("sources_get", pars, token=token)

    def get_report(self, rpt_url: str | KleioFile, token=None) -> str:
        """Get report from kleio server

        Args:

            rpt_url (str | KleioFile): report file url or KleioFile object
            token (str, optional): Kleio server token; defaults to None -> use admin token.

            Returns:
                str: report content
        """
        if isinstance(rpt_url, KleioFile):
            rpt_url = rpt_url.rpt_url

        if not rpt_url.startswith("/"):
            rpt_url = f"/{rpt_url}"

        server_url = f"{self.get_url()}{rpt_url}"
        return self.get_url_content(server_url, token=token)

    def get_source(self, src: str | KleioFile, token=None) -> str:
        """ Get the texto of a source

        Args:
            src (str | KleioFile): source file url or KleioFile object
            token (str, optional): Kleio server token; defaults to None -> use admin token.

        Returns:
            str: source text
        """
        if isinstance(src, KleioFile):
            url = src.source_url

        if not url.startswith("/"):
            url = f"/{url}"

        server_url = f"{self.get_url()}{url}"
        return self.get_url_content(server_url, token=token)

    def get_url_content(self, server_url: str, token=None, timeout=30) -> str:
        """Get content from Kleio Server

        Args:
            server_url (str): url of the content in the Kleio Server
            token (str, optional): Kleio server token; defaults to None -> use admin token.
            timeout (int, optional): timeout in seconds; defaults to 30.

        """
        if token is None:
            token = self.get_token()
        headers = {"Authorization": f"Bearer {token}"}
        req = urllib.request.Request(server_url, headers=headers)
        with urllib.request.urlopen(req, timeout=timeout) as source:
            response_content = source.read().decode("utf-8")
            return response_content

    def get_home_page(self, token=None) -> str:
        """Get home page from Kleio Server

        Args:
            server_url (str): url of the content in the Kleio Server

        """
        if token is None:
            token = self.get_token()
        if token is None:
            headers = {}
        else:
            headers = {"Authorization": f"Bearer {token}"}
        req = urllib.request.Request(self.get_url(), headers=headers)
        with urllib.request.urlopen(req) as source:
            response_content = source.read().decode("utf-8")
            return response_content

    def extract_version_info(self, home_page_content: str) -> Optional[Tuple[str, str, str]]:
        """Extract version information from the home page content.

        Args:
            home_page_content (str): The content of the home page.

        Returns:
            Optional[Tuple[str, str, str]]: A tuple containing the version number, build number, and date of build,
                                            or None if the information could not be extracted.
        """
        pattern = r"Version: KleioTranslator - server version (\d+\.\d+) - build (\d+) (\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})"
        match = re.search(pattern, home_page_content)
        if match:
            version = match.group(1)
            build_number = match.group(2)
            build_date = match.group(3)
            return version, build_number, build_date
        return None

    def get_version_info(self) -> Optional[Tuple[str, str, str]]:
        """Fetch the home page and extract version information.

        Returns:
            Optional[Tuple[str, str, str]]: A tuple containing the version number, build number, and date of build,
                                            or None if the information could not be extracted.
        """
        home_page_content = self.get_home_page()
        return self.extract_version_info(home_page_content)

    def __repr__(self):
        return (f"KleioServer(url={self.get_url()}, kleio_home={self.get_kleio_home()})")


def is_docker_running():
    """Check if docker is running"""
    try:
        client = docker.from_env()
        client.ping()
        return True
    except Exception as exec:
        logging.error(f"Could not connect to Docker. Is it running? {exec}")
        return False


def find_local_kleio_home(path: str = None):
    """Find kleio home directory.

    Kleio home directory is the directory where Kleio Server finds sources and auxiliary
    files like structures, mappings and inferences.

    It can be in the current directory, parent directory, or tests directory.
    It can be named "kleio-home", "timelink-home", or "mhk-home".

    Special cases:
       * if the current directory is "notebooks", kleio-home is assumed that Kleio home
            to be the parent directory of "notebooks"
       * if there is a "tests" subdirectory, kleio-home is searched in childs of "tests"
    """
    kleio_home = None
    timelink_home_names = ["kleio-home", "timelink-home", "mhk-home"]

    if path is None:
        # get the current directory
        current_dir = os.getcwd()
    else:
        current_dir = path

    # get the user home directory
    user_home = os.path.expanduser("~")

    # check if current_dir is "notebooks"
    if os.path.basename(current_dir) == "notebooks":
        kleio_home = os.path.dirname(current_dir)
    # check if current dir is one of the timelink home names
    elif os.path.basename(current_dir) in timelink_home_names:
        kleio_home = os.path.dirname(current_dir)
    else:
        # check if kleio-home exists in current directory,
        # parents of current directory up to user_home,
        # or tests sub directory of current directory
        dir_path = current_dir
        while dir_path != user_home:
            for home_dir in timelink_home_names:
                if os.path.isdir(f"{dir_path}/{home_dir}"):
                    kleio_home = f"{dir_path}/{home_dir}"
                    break
            if kleio_home:
                break
            dir_path = os.path.dirname(dir_path)
            if dir_path == os.path.dirname(dir_path):
                break  # we reached the root
        if kleio_home is None:
            # check if current_dir is "tests"
            if os.path.isdir(f"{current_dir}/tests"):
                dir_path = f"{current_dir}/tests"
                # check if kleio-home exists in tests directory
                for home_dir in timelink_home_names:
                    if os.path.isdir(f"{dir_path}/{home_dir}"):
                        kleio_home = f"{dir_path}/{home_dir}"
                        break
    if kleio_home is None:
        kleio_home = current_dir
    return kleio_home


def get_kserver_home(
    container: docker.models.containers.Container = None,
    container_number: int = 0,
    kleio_version="latest",
):
    """Get the kleio server home directory

    Args:
        container (docker.models.containers.Container, optional): kleio server container;
                                                         defaults to None -> get by number.
        container_number (int, optional): container number. Defaults to 0.
        kleio_version (str, optional): kleio server version; defaults to "latest"

    Returns the volume mapped to /kleio-home in the kleio server container"""

    if container is None:
        container = get_kserver_container_list(kleio_version=kleio_version)[
            container_number
        ]

    kleio_home = None
    if container is not None:
        kleio_home_mount = [
            mount["Source"]
            for mount in container.attrs["Mounts"]
            if mount["Destination"] == "/kleio-home"
        ]
        if len(kleio_home_mount) > 0:
            kleio_home = kleio_home_mount[0]
        else:
            kleio_home = None
    return kleio_home


def list_kleio_server_containers(
    kleio_version: str = None,
) -> List[Tuple[str, str, int, docker.models.containers.Container]]:
    """List running kleio server containers

    Args:
        kleio_version (str, optional): kleio server version; defaults
                                    to None -> any version.
    Returns:
        a tuple (name, kleio_home, port, token, container) for each container
    """
    containers: list[docker.models.containers.Container] = get_kserver_container_list(
        kleio_version=kleio_version
    )

    if containers is None or len(containers) == 0:
        return []

    result = []
    for container in containers:
        kleio_home = get_kserver_home(container)
        p = list(container.attrs["HostConfig"]["PortBindings"])[0]
        port = container.attrs["HostConfig"]["PortBindings"][p][0]["HostPort"]
        token = get_kserver_token(container)
        result.append((container.name, kleio_home, port, token, container))
    return result


def get_kserver_container(
    kleio_home: str = None, kleio_version="latest", stop_duplicates=False
):
    """Check if a kleio server is running in docker, possibly mapped to
    a given kleio home directory.

    Args:
        kleio_home (str, optional): kleio home directory; defaults to None -> any kleio home.
        kleio_version (str, optional): kleio server version; defaults to "latest"
                                        if not specified only checks latest
        stop_duplicates (bool, optional): if True, stop and remove duplicate containers
                                        with same kleio_home; defaults to False.
    Returns:
        docker.models.containers.Container: the Kleio server container
    """

    containers: list[docker.models.containers.Container] = get_kserver_container_list(
        kleio_version=kleio_version
    )

    if containers is None:
        return None
    elif kleio_home is not None:
        found = False
        kleio_home = os.path.abspath(kleio_home)
        for container in containers:
            kleio_home_mount = [
                mount["Source"]
                for mount in container.attrs["Mounts"]
                if mount["Destination"] == "/kleio-home"
            ]
            if kleio_home_mount[0] == kleio_home:
                if not found:
                    first_found = container
                    found = True
                else:
                    if stop_duplicates:
                        container.stop()
                        container.remove()
                    else:
                        break

        if not found:
            return None
        else:
            return first_found
    else:
        return containers[0]


def get_kserver_container_list(
    kleio_version: str = None,
) -> None | List[docker.models.containers.Container]:
    """Get the Kleio server containers currently running in docker

    Running containers are inspected to detect
    those based on images with "kleio-server" on a tag.

    Args:
        kleio_version (str, optional): kleio server version; defaults to "latest"
    Returns:
        docker.models.containers.Container: the Kleio server container
    """
    if is_docker_running() is False:
        raise Exception("Docker is not running")

    client: docker.DockerClient = docker.from_env()

    allcontainers = client.containers.list(all=False)
    allimages_tags = [(t, c) for c in allcontainers for t in c.image.tags]

    if kleio_version is None:
        containers = {c for t, c in allimages_tags if "kleio-server" in t}
    else:
        containers = {
            c for t, c in allimages_tags if f"kleio-server:{kleio_version}" in t
        }

    if len(containers) > 0:
        return list(containers)
    else:
        return None


def get_kserver_token(
    container: docker.models.containers.Container = None,
    container_number: int = 0,
    kleio_version="latest",
) -> str:
    """Get the Kleio server container admin token

    Args:
        container (docker.models.containers.Container, optional): kleio server container;
                                            defaults to None -> get the first container.
        container_number (int, optional): container number. Defaults to 0.
        kleio_version (str, optional): kleio server version; defaults to "latest"

    Returns:
        str: the kleio server container token
    """
    if container is None:
        container = get_kserver_container_list(kleio_version=kleio_version)[
            container_number
        ]

    token = [
        env
        for env in container.attrs["Config"]["Env"]
        if env.startswith("KLEIO_ADMIN_TOKEN")
    ][0].split("=")[1]
    return token


def find_free_port(from_port: int = 8088, to_port: int = 8099):
    for port in range(from_port, to_port + 1):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                pass
    raise OSError("No free ports available in the range 8088-8099")


def start_kleio_server(
    image: str = "timelinkserver/kleio-server",
    version: str | None = None,
    kleio_home: str | None = None,
    kleio_admin_token: str | None = None,
    kleio_server_port="8088",
    kleio_external_port=None,
    kleio_server_workers="3",
    kleio_idle_timeout=900,
    kleio_conf_dir=None,
    kleio_source_dir=None,
    kleio_stru_dir=None,
    kleio_token_db=None,
    kleio_default_stru=None,
    kleio_debug=None,
    consistency: str = "cached",
    update: bool = False,
    reuse: bool = True,
    stop_duplicates: bool = False,
):
    """Starts a kleio server in docker

    Args:
        image (str, optional): kleio server image. Defaults to "time-link/kleio-server".
        version (str | None, optional): kleio-server version; defaults to "latest".
        kleio_home (str | None, optional): kleio home directory. Defaults to current directory.
        kleio_admin_token (str | None, optional): kleio server token; if None -> generate a random token.
        kleio_server_port (str, optional): kleio server port (in the container). Defaults to "8088".
        kleio_external_port (str, optional): kleio server external port (in the host).
        kleio_server_workers (str, optional): number kleio server workers. Defaults to "3".
        kleio_idle_timeout (int, optional): kleio server idle timeout. Defaults to 900.
        kleio_conf_dir (str, optional): kleio server configuration directory. Defaults to None.
        kleio_source_dir (str, optional): kleio server sources directory. Defaults to None.
        kleio_stru_dir (str, optional): kleio server structures directory. Defaults to None.
        kleio_token_db (str, optional): kleio server token database. Defaults to None.
        kleio_default_stru (str, optional): kleio server default structure. Defaults to None.
        kleio_debug (str, optional): kleio server debug level. Defaults to None.

        consistency (str, optional): consistency of the volume mount. Defaults to "cached".
        port (int, optional): port to map to 8088, if none find a free port starting at 8088.
        update (bool, optional): update kleio server image. Defaults to False.
        reuse (bool, optional): if True, reuse an existing kleio server container
                                    with same keio_home; defaults to True.
        stop_duplicates (bool, optional): if True, stop and remove duplicate containers

    Usage:
        update=True will pull the latest image; if it is newer than existing one
        any container running with the same kleio_home will be stopped and removed,
        a new container will be started with the new image and reuse=True will be ignored.

    """
    # check if kleio server is already running in docker
    if is_docker_running() is False:
        raise RuntimeError("Docker is not running")

    exists = get_kserver_container(
        kleio_home=kleio_home, kleio_version=version, stop_duplicates=stop_duplicates
    )

    client = docker.from_env()
    if update is True:
        if version is None:
            get_version = "latest"
        else:
            get_version = version

        # Pull the latest image
        logging.info(
            "Pulling the latest image: timelinkserver/kleio-server:%s", get_version
        )
        latest_image = client.images.pull(
            "timelinkserver/kleio-server", tag=get_version
        )

        # Get the currently used image
        current_image = client.images.get(f"timelinkserver/kleio-server:{get_version}")

        # Compare the IDs
        if latest_image.id != current_image.id:
            logging.info("A newer version is available.")
            if exists is not None:
                logging.info("Current container was stopped and removed.")
                exists.stop()
                exists.remove()
                exists = None

        else:
            logging.info("You are using Kleio Server %s (updated)", get_version)

    if exists is not None:
        if reuse:
            logging.warning(f"Reusing existing container {exists.name}")
            return exists

    # if kleio_home is None, use current directory
    if kleio_home is None:
        kleio_home = os.getcwd()
    else:
        kleio_home = os.path.abspath(kleio_home)
        # check if dir exists
        if not os.path.exists(kleio_home):
            raise FileNotFoundError(f"Directory {kleio_home} does not exist")

    # ensure that kleio_home/system/conf/kleio exists
    # TODO: remove this
    # os.makedirs(f"{kleio_home}/system/conf/kleio", exist_ok=True)

    if kleio_admin_token is None:
        kleio_admin_token = random_token()

    if kleio_external_port is None:
        kleio_external_port = find_free_port(8088, 8099)

    kleio_env = dict()
    if kleio_conf_dir is not None:
        kleio_env["KLEIO_CONF_DIR"] = kleio_conf_dir
    if kleio_source_dir is not None:
        kleio_env["KLEIO_SOURCE_DIR"] = kleio_source_dir
    if kleio_stru_dir is not None:
        kleio_env["KLEIO_STRU_DIR"] = kleio_stru_dir
    if kleio_token_db is not None:
        kleio_env["KLEIO_TOKEN_DB"] = kleio_token_db
    if kleio_default_stru is not None:
        kleio_env["KLEIO_DEFAULT_STRU"] = kleio_default_stru
    if kleio_debug is not None:
        kleio_env["KLEIO_DEBUG"] = kleio_debug
    if kleio_server_workers is not None:
        kleio_env["KLEIO_SERVER_WORKERS"] = kleio_server_workers
    if kleio_idle_timeout is not None:
        kleio_env["KLEIO_IDLE_TIMEOUT"] = kleio_idle_timeout
    if kleio_admin_token is not None:
        kleio_env["KLEIO_ADMIN_TOKEN"] = kleio_admin_token
    if kleio_home is not None:
        kleio_env["KLEIO_HOME"] = kleio_home
    if kleio_server_port is not None:
        kleio_env["KLEIO_SERVER_PORT"] = kleio_server_port

    kleio_container = client.containers.run(
        image=f"{image}:{version}",
        detach=True,
        ports={f"{kleio_server_port}/tcp": kleio_external_port},
        environment=kleio_env,
        mounts=[
            docker.types.Mount(
                target="/kleio-home",
                source=kleio_home,
                type="bind",
                read_only=False,
                consistency=consistency,
            )
        ],
    )

    timeout = 15
    stop_time = 1
    elapsed_time = 0
    # this necessary to get the status
    cont = client.containers.get(kleio_container.id)
    while cont.status not in ["running"] and elapsed_time < timeout:
        sleep(stop_time)
        cont = client.containers.get(kleio_container.id)
        elapsed_time += stop_time
    if cont.status != "running":
        raise RuntimeError("Kleio server did not start")

    return kleio_container


def random_token(length=32):
    """Generate a random token"""
    alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return "".join(secrets.choice(alphabet) for i in range(length))


def stop_kleio_server(container: docker.models.containers.Container = None):
    """Stop kleio server"""
    if is_docker_running() is False:
        raise Exception("Docker is not running")

    if container is None:
        container = get_kserver_container()

    container = get_kserver_container()
    container.stop()
    container.remove()



"""
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""

from sqlalchemy import Column, String, ForeignKey

from timelink.kleio.utilities import quote_long_text, render_with_extra_info
from .entity import Entity


class Geoentity(Entity):
    """ represents a geographical entity."""

    __tablename__ = "geoentities"

    id = Column(String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True)
    name = Column(String, index=True)
    the_type = Column(String(32), nullable=True)
    obs = Column(String)

    __mapper_args__ = {"polymorphic_identity": "geoentity"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Geoentity(id={sr}, "
            f'name="{self.name}", '
            f'the_type="{self.the_type}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        return self.to_kleio(show_contained=False)

    def to_kleio(self, self_string=None, show_contained=False, ident="", ident_inc="  ", width=80, **kwargs) -> str:

        obs, extra_info = self.get_extra_info()
        if self.groupname is None:
            myname = "attribute"
        else:
            myname = self.groupname
            r = f"{myname}${render_with_extra_info('name', self.name, extra_info=extra_info, **kwargs)}"
            r += f"/{render_with_extra_info('the_type', self.the_type, extra_info=extra_info, **kwargs)}"
            if obs is not None and len(obs.strip()) > 0:
                self_string = f"{r}/obs={quote_long_text(obs.strip(), width=width)}"
            else:
                self_string = r
        r = super().to_kleio(self_string=self_string, ident=ident, ident_inc=ident_inc,
                             show_contained=show_contained, width=width, **kwargs)
        return r



import os
import sys
import time
from datetime import datetime

from pathlib import Path
from typing import Union
import platform
import urllib.request

from sqlalchemy.orm import Session
from sqlalchemy import func
from xml.sax import make_parser

from .sax_handler import SaxHandler
from .kleio_handler import KleioHandler


def import_from_xml(
    filespec: Union[str, Path], session: Session, options: dict = None
) -> dict:
    """Import data from file or url into a timelink-mhk database.

    The data file must be a XML file in the format exported by the ``kleio``
    translator.

    Arguments:
        filespec (str,Path): a file path, URL of a data file or path in a Kleio Server.
        session: a database session from TimelinkDatabase.session()
        options (dict): a dictionnary with options

           - 'return_stats':  if True import stats will be returned
           - 'kleio_url':  the url of kleio server;
           - 'kleio_token':  the authorization token for the kleio server.
           - 'mode':  the mode of the import, either 'TL'(Timelink) or 'MHK'

        If kleio_url and kleio_token are specified the data will be fetched from
        a KleioServer and the filespec should contain the "xml_path" of the file
        in the server (e.g. /rest/exports/reference_sources/varia/vereacao.xml) and
        the kleio_token will be inserted in the header for autentication

    Returns:
        If stats is True in options a dict with statistical information will be
        returned.

            - 'datetime': the time stamp of the start of the import
            - 'machine':  local machine name
            - 'file': path to imported file or url
            - 'import_time_seconds':  elapsed time during import
            - 'entities_processed': number of entities processed
            - 'entity_rate': number of entities processed per second
            - 'person_rate': number of persons (entities of class 'person')
            - 'nerrors': number of errors during import
            - 'errors': list of error messages

    Examples:
        Returned statistical information when stats=True
    ::

        {
        'datetime': '2022-01-02 18:11:12',
        'machine': 'joaquims-mbpr.local',
        'file': 'https://...b1685.xml',
        'import_time_seconds': 7.022288084030151,
        'entities_processed': 747,
        'entity_rate': 106.37558457603042,
        'person_rate': 27.483919441999827
        'nerrors': 0
        'errors': []
        }

        TODO: should use https when the kleio_url not local.
    """
    collect_stats = False
    kleio_url = None
    kleio_token = None
    mode = "TL"  # determines the database model TL=Timelink, MHK=MHK
    nentities_before = 0
    npersons_before = 0
    now = datetime.now()
    if options is not None and options.get("return_stats", False):
        collect_stats = True
    if options is not None and options.get("mode", None) is not None:
        mode = options.get("mode")
    if options is not None and options.get("kleio_url", None) is not None:
        kleio_url = options.get("kleio_url")
    if options is not None and options.get("kleio_token", None) is not None:
        kleio_token = options.get("kleio_token")

    kleio_handler = KleioHandler(session, mode=mode)
    sax_handler = SaxHandler(kleio_handler)
    parser = make_parser()
    parser.setContentHandler(sax_handler)
    start = time.time()
    if collect_stats:
        nentities_before = session.query(
            func.count(kleio_handler.entity_model.id)  # pylint: disable=not-callable
        ).scalar()
        npersons_before = session.query(
            func.count(kleio_handler.person_model.id)  # pylint: disable=not-callable
        ).scalar()

    if kleio_url is not None and kleio_token is not None:
        headers = {"Authorization": f"Bearer {kleio_token}"}
        server_url = f"{kleio_url}{filespec}"
        req = urllib.request.Request(server_url, headers=headers)
        with urllib.request.urlopen(req, timeout=30) as source:
            parser.parse(source)
    elif kleio_token is not None or kleio_url is not None:
        # this means that one of the options is missing
        raise ValueError(
            "Both kleio_url and kleio_token must be specified to fetch from kleio server"
        )
    elif isinstance(filespec, os.PathLike):
        source = os.fspath(filespec)
        parser.parse(source)
    else:
        source = filespec
        parser.parse(source)

    end = time.time()
    if collect_stats:
        machine = platform.node()
        nentities = session.query(func.count(kleio_handler.entity_model.id)).scalar()  # pylint: disable=not-callable
        npersons = session.query(func.count(kleio_handler.person_model.id)).scalar()  # pylint: disable=not-callable
        erate = (nentities - nentities_before) / (end - start)
        prate = (npersons - npersons_before) / (end - start)
        stats = {
            "datetime": now.timestamp(),
            "machine": machine,
            "database": session.bind.url,
            "file": filespec,
            "import_time_seconds": end - start,
            "entities_processed": nentities - nentities_before,
            "entity_rate": erate,
            "person_rate": prate,
            "nerrors": len(kleio_handler.errors),
            "errors": kleio_handler.errors,
        }
        return stats


if (
    __name__ == "__main__"
):  # Not sure this works, where is the session for KleioHandler?
    saxParser = make_parser()
    kleioHandler = KleioHandler(None, mode="TL")
    saxParser.setContentHandler(SaxHandler(kleioHandler))
    saxParser.parse(sys.argv[1])



import warnings

# pylint: disable=import-error
from sqlalchemy import Column, String, ForeignKey, Index
from sqlalchemy.orm import relationship

from timelink.kleio.utilities import (
    kleio_escape,
    quote_long_text,
    format_timelink_date as ftld,
)
from .entity import Entity


class Relation(Entity):
    """represents a relation between two entities.

    Relations have a type, a value, a date and an optional observation.

    Args:
        id (str): the id of the relation
        origin (str): the id of the origin entity
        destination (str): the id of the destination entity
        the_type (str): the type of the relation
        the_value (str): the value of the relation
        the_date (str): the date of the relation
        obs (str): an observation about the relation

    Also, for the entity super class:
        groupname (str): the name of the group
        inside (str): the id of the containing entity
        the_line (str): the line of the entity in the source
        the_level (str): the level of the entity in the source
        the_order (str): the order of the entity in the source

    """

    __tablename__ = "relations"

    id = Column(String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True)
    # rel_entity = relationship("Entity",
    #                   foreign_keys='id',back_populates='rel')
    origin = Column(String, ForeignKey("entities.id"), index=True)
    org = relationship(Entity, foreign_keys=[origin], back_populates="rels_out")

    destination = Column(String, ForeignKey("entities.id"), index=True)
    dest = relationship("Entity", foreign_keys=[destination], back_populates="rels_in")
    the_type = Column(String, index=True)
    the_value = Column(String, index=True)
    the_date = Column(String, index=True)
    obs = Column(String)

    __mapper_args__ = {
        "polymorphic_identity": "relation",
        "inherit_condition": id == Entity.id,
    }

    __table_args__ = (Index("relations_type_value", "the_type", "the_value"),)

    @property
    def dest_class(self):
        if self.dest is None:
            relinfo = repr(self)
            warnings.warn(f"Missing destination for relation\n {relinfo}", UserWarning, stacklevel=2)
            return None
        return self.dest.pom_class

    @property
    def org_class(self):
        return self.org.pom_class

    @property
    def dest_name(self):
        if self.dest_class is None:
            return ""
        elif self.dest_class == "person" or self.dest_class == "object":
            return self.dest.name
        else:
            return self.dest.groupname

    @property
    def org_name(self):
        if self.org_class == "person" or self.org_class == "object":
            return self.org.name
        else:
            return self.org.groupname

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Relation(id={sr}, "
            f'origin="{self.origin}", '
            f'destination="{self.destination}", '
            f'the_type="{self.the_type}", '
            f'the_value="{self.the_value}", '
            f'the_date="{self.the_date}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        return self.to_kleio()

    def to_kleio(self, ident="", **kwargs):  # pylint: disable=arguments-differ

        outgoing = kwargs.get("outgoing", True)
        if outgoing:
            if self.the_type == "function-in-act":
                function = self.the_value
                act_type = self.dest.groupname
                act_date = self.dest.the_date
                act_id = self.dest.id

                s = f"{ident}rel$function-in-act/{function}/{act_type}/{act_id}/{ftld(act_date)}"
                if self.the_line is not None:
                    s += f"/obs=line: {self.the_line}"
                return s

            if self.the_type == "identification":

                return f"{ident}rel$identification/{self.the_value}/{self.dest_name}/{self.destination}/{ftld(self.the_date)}"

            if self.dest is not None:
                if self.dest.pom_class in [
                    "person",
                    "object",
                    "geoentity",
                ]:
                    relname = self.dest.name
                else:
                    relname = self.dest.groupname
            else:
                relname = "*missing*"
            label = "rel"
        else:  # incoming relation

            if self.org is not None and self.org.pom_class in [
                "person",
                "object",
                "geoentity",
            ]:
                relname = self.org.name
            else:
                relname = self.org.groupname
            label = "<rel"

        r = (
            f"{ident}{label}${kleio_escape(self.the_type)}/"
            f"{quote_long_text(self.the_value)}/"
            f"{kleio_escape(relname)}/{self.origin}/{ftld(self.the_date)}"
        )
        if self.obs is not None and len(self.obs.strip()) > 0:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r


Entity.rels_out = relationship(
    "Relation", foreign_keys=[Relation.origin], back_populates="dest"
)

Entity.rels_in = relationship(
    "Relation", foreign_keys=[Relation.destination], back_populates="org"
)



from typing import Annotated
from fastapi import Request, Header
from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from fastapi import HTTPException
from jose import JWTError
from timelink.app.models.user_database import UserDatabase
from timelink.app.schemas import UserSchema


from timelink.app.services.auth import fief, auth, FiefUserInfo, SESSION_COOKIE_NAME
from timelink.app.models.user import User

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


def get_fastui_token(authorization: Annotated[str, Header()] = ''):
    try:
        token = authorization.split(' ', 1)[1]
    except IndexError:
        return None
    return token


async def get_current_user(
    request: Request, user: Annotated[FiefUserInfo, Depends(auth.current_user(optional=True))]
):
    try:
        users_db: UserDatabase = request.app.state.webapp.users_db
        if user is None:
            return None
        else:
            user_schema = users_db.cache.get(user["email"])
            if user_schema is None:
                with users_db.session() as session:
                    cookie = request.cookies.get(SESSION_COOKIE_NAME)
                    token_access = await fief.validate_access_token(cookie)
                    permissions = token_access.get("permissions", [])
                    db_user = users_db.on_board_user(user, permissions=permissions, session=session)
                    user_schema = UserSchema.model_validate(db_user)
                    user_schema.properties = users_db.get_user_properties(db_user.id, session=session)
                    user_schema.projects = users_db.get_user_projects(db_user.id, session=session)
                    users_db.cache[user["email"]] = user_schema
            return user_schema

    except JWTError as jexception:
        raise jexception
        # raise credentials_exception

    return user_schema


async def get_current_active_user(
    request: Request, current_user: Annotated[UserSchema, Depends(get_current_user)]
):
    if current_user.disabled:
        raise HTTPException(status_code=400, detail="Inactive user")
    return current_user


# Dependency to get a connection to the database
def get_user_db(request: Request):
    """Get the User database request"""

    webapp = request.app.webapp
    return webapp.users_db


# dependency to get a connection to the kleio server
def get_kleio_server(request: Request):
    """Get a connection to the kleio server

    Uses timelink.kleio.kleio_server.KleioServer to get a connection to the kleio server.
    """
    webapp = request.app.webapp

    return webapp.kleio_server


def get_db(request: Request):
    """Get the database request"""
    webapp = request.app.webapp
    user = Annotated[User, Depends(get_current_active_user)]
    # Todo: Determine the user's project
    # project = webapp.projects[user.name]
    # project_db = project.db
    return webapp.get_current_project_db(user)


def get_github_auth(request: Request):
    """Get the github auth request"""
    webapp = request.app.webapp
    return webapp



# pylint: disable=import-error

from sqlalchemy import Column, String, ForeignKey, Index
from sqlalchemy.orm import relationship

from timelink.kleio.utilities import quote_long_text
from timelink.mhk.models.entity import Entity


class Attribute(Entity):
    __table_args__ = {"extend_existing": True}
    __tablename__ = "attributes"

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    entity = Column(String, ForeignKey("entities.id"), index=True)
    the_type = Column(String, index=True)
    the_value = Column(String, index=True)
    the_date = Column(String, index=True)
    obs = Column(String)
    the_entity = relationship(
        "Entity",
        foreign_keys=[entity],
        back_populates="attributes",
    )

    __mapper_args__ = {
        "polymorphic_identity": "attribute",
        "inherit_condition": id == Entity.id,
    }

    __table_args__ = (Index("attributes_type_value", "the_type", "the_value"),)

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Attribute(id={sr}, "
            f'entity="{self.entity}", '
            f'the_type="{self.the_type}", '
            f'the_value="{self.the_value}", '
            f'the_date="{self.the_date}"", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        r = f"{self.groupname}${quote_long_text(self.the_type)}/{quote_long_text(self.the_value)}/"
        r += f"{self.the_date}"
        if self.obs is not None and len(self.obs.strip()) > 0:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r


Entity.attributes = relationship(
    "Attribute", foreign_keys=[Attribute.entity], back_populates="the_entity"
)



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kact import KAct


class KPerson(KGroup):
    """KPerson(name,sex,id,obs=,same_as=,xsame_as=)

    Person in a historical source

    Elements:
        name: the name of the person. A string.
        sex:  the gender of the person. A string.
        id: an unique id for this person. A string, optional.
        obs: a note on the person. A string, optional.
        same_as: id of another instance of this person in the same file.
        xsame_as: id of another instance of this person in another file.

    Kleio str definition:

    part	name=person ;
            guaranteed=name,sex;
            also=id,obs,same_as;
            position=name,sex,id,same_as;
            arbitrary=atr,rel,ls
    """

    _name = "person"
    _guaranteed = ["name", "sex"]
    _also = ["id", "obs", "same_as"]
    _position = ["name", "sex", "id", "same_as", "xsame_as"]
    _part = ["rel", "attr"]
    _pom_class_id: str = "person"


KAct.allow_as_part(KPerson)
